from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.borsh_extension import EnumForCodegen


class SOLJSON(typing.TypedDict):
    kind: typing.Literal["SOL"]


class BTCJSON(typing.TypedDict):
    kind: typing.Literal["BTC"]


class ETHJSON(typing.TypedDict):
    kind: typing.Literal["ETH"]


class APTJSON(typing.TypedDict):
    kind: typing.Literal["APT"]


class ARBJSON(typing.TypedDict):
    kind: typing.Literal["ARB"]


class BNBJSON(typing.TypedDict):
    kind: typing.Literal["BNB"]


class PYTHJSON(typing.TypedDict):
    kind: typing.Literal["PYTH"]


class TIAJSON(typing.TypedDict):
    kind: typing.Literal["TIA"]


class JTOJSON(typing.TypedDict):
    kind: typing.Literal["JTO"]


class ONEMBONKJSON(typing.TypedDict):
    kind: typing.Literal["ONEMBONK"]


class SEIJSON(typing.TypedDict):
    kind: typing.Literal["SEI"]


class JUPJSON(typing.TypedDict):
    kind: typing.Literal["JUP"]


class DYMJSON(typing.TypedDict):
    kind: typing.Literal["DYM"]


class STRKJSON(typing.TypedDict):
    kind: typing.Literal["STRK"]


class WIFJSON(typing.TypedDict):
    kind: typing.Literal["WIF"]


class RNDRJSON(typing.TypedDict):
    kind: typing.Literal["RNDR"]


class TNSRJSON(typing.TypedDict):
    kind: typing.Literal["TNSR"]


class POPCATJSON(typing.TypedDict):
    kind: typing.Literal["POPCAT"]


class EIGENJSON(typing.TypedDict):
    kind: typing.Literal["EIGEN"]


class DBRJSON(typing.TypedDict):
    kind: typing.Literal["DBR"]


class GOATJSON(typing.TypedDict):
    kind: typing.Literal["GOAT"]


class UNDEFINEDJSON(typing.TypedDict):
    kind: typing.Literal["UNDEFINED"]


@dataclass
class SOL:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "SOL"

    @classmethod
    def to_json(cls) -> SOLJSON:
        return SOLJSON(
            kind="SOL",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "SOL": {},
        }


@dataclass
class BTC:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "BTC"

    @classmethod
    def to_json(cls) -> BTCJSON:
        return BTCJSON(
            kind="BTC",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "BTC": {},
        }


@dataclass
class ETH:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "ETH"

    @classmethod
    def to_json(cls) -> ETHJSON:
        return ETHJSON(
            kind="ETH",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "ETH": {},
        }


@dataclass
class APT:
    discriminator: typing.ClassVar = 3
    kind: typing.ClassVar = "APT"

    @classmethod
    def to_json(cls) -> APTJSON:
        return APTJSON(
            kind="APT",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "APT": {},
        }


@dataclass
class ARB:
    discriminator: typing.ClassVar = 4
    kind: typing.ClassVar = "ARB"

    @classmethod
    def to_json(cls) -> ARBJSON:
        return ARBJSON(
            kind="ARB",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "ARB": {},
        }


@dataclass
class BNB:
    discriminator: typing.ClassVar = 5
    kind: typing.ClassVar = "BNB"

    @classmethod
    def to_json(cls) -> BNBJSON:
        return BNBJSON(
            kind="BNB",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "BNB": {},
        }


@dataclass
class PYTH:
    discriminator: typing.ClassVar = 6
    kind: typing.ClassVar = "PYTH"

    @classmethod
    def to_json(cls) -> PYTHJSON:
        return PYTHJSON(
            kind="PYTH",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "PYTH": {},
        }


@dataclass
class TIA:
    discriminator: typing.ClassVar = 7
    kind: typing.ClassVar = "TIA"

    @classmethod
    def to_json(cls) -> TIAJSON:
        return TIAJSON(
            kind="TIA",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "TIA": {},
        }


@dataclass
class JTO:
    discriminator: typing.ClassVar = 8
    kind: typing.ClassVar = "JTO"

    @classmethod
    def to_json(cls) -> JTOJSON:
        return JTOJSON(
            kind="JTO",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "JTO": {},
        }


@dataclass
class ONEMBONK:
    discriminator: typing.ClassVar = 9
    kind: typing.ClassVar = "ONEMBONK"

    @classmethod
    def to_json(cls) -> ONEMBONKJSON:
        return ONEMBONKJSON(
            kind="ONEMBONK",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "ONEMBONK": {},
        }


@dataclass
class SEI:
    discriminator: typing.ClassVar = 10
    kind: typing.ClassVar = "SEI"

    @classmethod
    def to_json(cls) -> SEIJSON:
        return SEIJSON(
            kind="SEI",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "SEI": {},
        }


@dataclass
class JUP:
    discriminator: typing.ClassVar = 11
    kind: typing.ClassVar = "JUP"

    @classmethod
    def to_json(cls) -> JUPJSON:
        return JUPJSON(
            kind="JUP",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "JUP": {},
        }


@dataclass
class DYM:
    discriminator: typing.ClassVar = 12
    kind: typing.ClassVar = "DYM"

    @classmethod
    def to_json(cls) -> DYMJSON:
        return DYMJSON(
            kind="DYM",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "DYM": {},
        }


@dataclass
class STRK:
    discriminator: typing.ClassVar = 13
    kind: typing.ClassVar = "STRK"

    @classmethod
    def to_json(cls) -> STRKJSON:
        return STRKJSON(
            kind="STRK",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "STRK": {},
        }


@dataclass
class WIF:
    discriminator: typing.ClassVar = 14
    kind: typing.ClassVar = "WIF"

    @classmethod
    def to_json(cls) -> WIFJSON:
        return WIFJSON(
            kind="WIF",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "WIF": {},
        }


@dataclass
class RNDR:
    discriminator: typing.ClassVar = 15
    kind: typing.ClassVar = "RNDR"

    @classmethod
    def to_json(cls) -> RNDRJSON:
        return RNDRJSON(
            kind="RNDR",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "RNDR": {},
        }


@dataclass
class TNSR:
    discriminator: typing.ClassVar = 16
    kind: typing.ClassVar = "TNSR"

    @classmethod
    def to_json(cls) -> TNSRJSON:
        return TNSRJSON(
            kind="TNSR",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "TNSR": {},
        }


@dataclass
class POPCAT:
    discriminator: typing.ClassVar = 17
    kind: typing.ClassVar = "POPCAT"

    @classmethod
    def to_json(cls) -> POPCATJSON:
        return POPCATJSON(
            kind="POPCAT",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "POPCAT": {},
        }


@dataclass
class EIGEN:
    discriminator: typing.ClassVar = 18
    kind: typing.ClassVar = "EIGEN"

    @classmethod
    def to_json(cls) -> EIGENJSON:
        return EIGENJSON(
            kind="EIGEN",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "EIGEN": {},
        }


@dataclass
class DBR:
    discriminator: typing.ClassVar = 19
    kind: typing.ClassVar = "DBR"

    @classmethod
    def to_json(cls) -> DBRJSON:
        return DBRJSON(
            kind="DBR",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "DBR": {},
        }


@dataclass
class GOAT:
    discriminator: typing.ClassVar = 20
    kind: typing.ClassVar = "GOAT"

    @classmethod
    def to_json(cls) -> GOATJSON:
        return GOATJSON(
            kind="GOAT",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "GOAT": {},
        }


@dataclass
class UNDEFINED:
    discriminator: typing.ClassVar = 21
    kind: typing.ClassVar = "UNDEFINED"

    @classmethod
    def to_json(cls) -> UNDEFINEDJSON:
        return UNDEFINEDJSON(
            kind="UNDEFINED",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "UNDEFINED": {},
        }


AssetKind = typing.Union[
    SOL,
    BTC,
    ETH,
    APT,
    ARB,
    BNB,
    PYTH,
    TIA,
    JTO,
    ONEMBONK,
    SEI,
    JUP,
    DYM,
    STRK,
    WIF,
    RNDR,
    TNSR,
    POPCAT,
    EIGEN,
    DBR,
    GOAT,
    UNDEFINED,
]
AssetJSON = typing.Union[
    SOLJSON,
    BTCJSON,
    ETHJSON,
    APTJSON,
    ARBJSON,
    BNBJSON,
    PYTHJSON,
    TIAJSON,
    JTOJSON,
    ONEMBONKJSON,
    SEIJSON,
    JUPJSON,
    DYMJSON,
    STRKJSON,
    WIFJSON,
    RNDRJSON,
    TNSRJSON,
    POPCATJSON,
    EIGENJSON,
    DBRJSON,
    GOATJSON,
    UNDEFINEDJSON,
]


def from_decoded(obj: dict) -> AssetKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "SOL" in obj:
        return SOL()
    if "BTC" in obj:
        return BTC()
    if "ETH" in obj:
        return ETH()
    if "APT" in obj:
        return APT()
    if "ARB" in obj:
        return ARB()
    if "BNB" in obj:
        return BNB()
    if "PYTH" in obj:
        return PYTH()
    if "TIA" in obj:
        return TIA()
    if "JTO" in obj:
        return JTO()
    if "ONEMBONK" in obj:
        return ONEMBONK()
    if "SEI" in obj:
        return SEI()
    if "JUP" in obj:
        return JUP()
    if "DYM" in obj:
        return DYM()
    if "STRK" in obj:
        return STRK()
    if "WIF" in obj:
        return WIF()
    if "RNDR" in obj:
        return RNDR()
    if "TNSR" in obj:
        return TNSR()
    if "POPCAT" in obj:
        return POPCAT()
    if "EIGEN" in obj:
        return EIGEN()
    if "DBR" in obj:
        return DBR()
    if "GOAT" in obj:
        return GOAT()
    if "UNDEFINED" in obj:
        return UNDEFINED()
    raise ValueError("Invalid enum object")


def from_json(obj: AssetJSON) -> AssetKind:
    if obj["kind"] == "SOL":
        return SOL()
    if obj["kind"] == "BTC":
        return BTC()
    if obj["kind"] == "ETH":
        return ETH()
    if obj["kind"] == "APT":
        return APT()
    if obj["kind"] == "ARB":
        return ARB()
    if obj["kind"] == "BNB":
        return BNB()
    if obj["kind"] == "PYTH":
        return PYTH()
    if obj["kind"] == "TIA":
        return TIA()
    if obj["kind"] == "JTO":
        return JTO()
    if obj["kind"] == "ONEMBONK":
        return ONEMBONK()
    if obj["kind"] == "SEI":
        return SEI()
    if obj["kind"] == "JUP":
        return JUP()
    if obj["kind"] == "DYM":
        return DYM()
    if obj["kind"] == "STRK":
        return STRK()
    if obj["kind"] == "WIF":
        return WIF()
    if obj["kind"] == "RNDR":
        return RNDR()
    if obj["kind"] == "TNSR":
        return TNSR()
    if obj["kind"] == "POPCAT":
        return POPCAT()
    if obj["kind"] == "EIGEN":
        return EIGEN()
    if obj["kind"] == "UNDEFINED":
        return UNDEFINED()
    if obj["kind"] == "DBR":
        return DBR()
    if obj["kind"] == "GOAT":
        return GOAT()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "SOL" / borsh.CStruct(),
    "BTC" / borsh.CStruct(),
    "ETH" / borsh.CStruct(),
    "APT" / borsh.CStruct(),
    "ARB" / borsh.CStruct(),
    "BNB" / borsh.CStruct(),
    "PYTH" / borsh.CStruct(),
    "TIA" / borsh.CStruct(),
    "JTO" / borsh.CStruct(),
    "ONEMBONK" / borsh.CStruct(),
    "SEI" / borsh.CStruct(),
    "JUP" / borsh.CStruct(),
    "DYM" / borsh.CStruct(),
    "STRK" / borsh.CStruct(),
    "WIF" / borsh.CStruct(),
    "RNDR" / borsh.CStruct(),
    "TNSR" / borsh.CStruct(),
    "POPCAT" / borsh.CStruct(),
    "EIGEN" / borsh.CStruct(),
    "DBR" / borsh.CStruct(),
    "GOAT" / borsh.CStruct(),
    "UNDEFINED" / borsh.CStruct(),
)



from __future__ import annotations

import logging
import os
from dataclasses import dataclass

from anchorpy import EventParser, Idl, Program, Provider, Wallet
from solana.rpc.async_api import AsyncClient
from solders.pubkey import Pubkey

from zetamarkets_py import constants, pda, utils
from zetamarkets_py.market import Market
from zetamarkets_py.types import Asset, Network
from zetamarkets_py.zeta_client.accounts.pricing import Pricing
from zetamarkets_py.zeta_client.accounts.state import State

idl_path = os.path.join(os.path.dirname(__file__), "idl/zeta.json")
with open(idl_path, "r") as f:
    idl = Idl.from_json(f.read())


@dataclass
class Exchange:
    """
    This class represents the Zeta Exchange. It contains all the necessary attributes and methods
    for interacting with the Zeta Exchange, including loading the exchange, fetching state and pricing,
    and handling various market assets.

    Note:
        Loading the exchange is asynchronous, so it is recommended to use :func:`load` to
        initialize the exchange.
    """

    # Initialize
    connection: AsyncClient
    """The connection to the Solana network."""
    program_id: Pubkey
    """The public key of the Zeta program."""
    program: Program
    """The Zeta program."""
    state: State
    """The state account of the Zeta program."""
    pricing: Pricing
    """The pricing account of the Zeta program. Stores mark prices, funding rates etc."""
    markets: dict[Asset, Market]
    """A dictionary mapping assets to their respective markets."""

    _event_parser: EventParser

    _state_address: Pubkey
    _pricing_address: Pubkey
    _serum_authority_address: Pubkey
    _mint_authority_address: Pubkey

    _logger: logging.Logger

    @classmethod
    async def load(
        cls,
        network: Network,
        connection: AsyncClient,
        assets: list[Asset] = Asset.all(),
        log_level: int = logging.CRITICAL,
    ) -> "Exchange":
        """
        Asynchronously load the Zeta Exchange.

        Args:
            network (Network): The network to connect to.
            connection (AsyncClient): The connection to the Solana network.
            assets (list[Asset], optional): The list of assets to load. Defaults to all assets.
            log_level (int, optional): The logging level. Defaults to logging.CRITICAL.

        Returns:
            Exchange: The loaded Zeta Exchange.

        Raises:
            Exception: If the state or pricing is not found at their respective addresses.
        """
        program_id = constants.ZETA_PID[network]
        provider = Provider(connection, Wallet.dummy())
        program = Program(idl, program_id, provider)
        _event_parser = EventParser(program_id, program.coder)

        # Accounts
        state_address = pda.get_state_address(program_id)
        state = await State.fetch(connection, state_address, connection.commitment, program_id=program_id)
        if state is None:
            raise Exception(f"State not found at {state_address}")

        pricing_address = pda.get_pricing_address(program_id)
        pricing = await Pricing.fetch(connection, pricing_address, connection.commitment, program_id=program_id)
        if pricing is None:
            raise Exception(f"Pricing not found at {pricing_address}")

        # Addresses
        _serum_authority_address = pda.get_serum_authority_address(program_id)
        _mint_authority_address = pda.get_mint_authority_address(program_id)

        markets = {
            asset: await Market.load(network, connection, asset, pricing.markets[asset.to_index()]) for asset in assets
        }

        # not currently used
        logger = utils.create_logger(f"{__name__}.{cls.__name__}", log_level)

        instance = cls(
            connection=connection,
            program_id=program_id,
            program=program,
            state=state,
            pricing=pricing,
            markets=markets,
            _event_parser=_event_parser,
            _state_address=state_address,
            _pricing_address=pricing_address,
            _serum_authority_address=_serum_authority_address,
            _mint_authority_address=_mint_authority_address,
            _logger=logger,
        )

        return instance

    @property
    def endpoint(self) -> str:
        """
        Returns the RPC endpoint.

        Returns:
            str: The endpoint URI.
        """
        return self.connection._provider.endpoint_uri

    @property
    def assets(self) -> list[Asset]:
        """
        Returns a list of keys from the markets dictionary.

        Returns:
            list[Asset]: A list of Asset objects.
        """
        return list(self.markets.keys())



from solders.pubkey import Pubkey
from spl.token.constants import ASSOCIATED_TOKEN_PROGRAM_ID

from zetamarkets_py.constants import FLEXIBLE_MINTS, MINTS, Asset
from zetamarkets_py.types import Network

TOKEN_PROGRAM_ID = Pubkey.from_string("TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA")


def get_state_address(program_id: Pubkey) -> Pubkey:
    return Pubkey.find_program_address([b"state"], program_id)[0]


def get_pricing_address(program_id: Pubkey) -> Pubkey:
    return Pubkey.find_program_address([b"pricing"], program_id)[0]


def get_zeta_group_address(program_id: Pubkey, mint: Pubkey) -> Pubkey:
    return Pubkey.find_program_address([b"zeta-group", bytes(mint)], program_id)[0]


def get_perp_sync_queue_address(program_id: Pubkey, zeta_group: Pubkey) -> Pubkey:
    return Pubkey.find_program_address([b"perp-sync-queue", bytes(zeta_group)], program_id)[0]


def get_margin_account_address(
    program_id: Pubkey,
    authority: Pubkey,
    subaccount_index: int = 0,
) -> Pubkey:
    return Pubkey.find_program_address(
        [
            b"cross-margin",
            bytes(authority),
            bytes([subaccount_index]),
        ],
        program_id,
    )[0]


def get_open_orders_address(
    program_id: Pubkey, dex_program_id: Pubkey, market: Pubkey, margin_account: Pubkey
) -> Pubkey:
    return Pubkey.find_program_address(
        [b"cross-open-orders", bytes(dex_program_id), bytes(market), bytes(margin_account)],
        program_id,
    )[0]


def get_cross_margin_account_manager_address(program_id: Pubkey, authority: Pubkey) -> Pubkey:
    return Pubkey.find_program_address(
        [b"cross-margin-manager", bytes(authority)],
        program_id,
    )[0]


def get_combined_vault_address(program_id: Pubkey) -> Pubkey:
    return Pubkey.find_program_address(
        [b"combined-vault"],
        program_id,
    )[0]


def get_zeta_vault_address(program_id: Pubkey, mint: Pubkey) -> Pubkey:
    return Pubkey.find_program_address(
        [b"zeta-vault", bytes(mint)],
        program_id,
    )[0]


def get_combined_socialized_loss_address(program_id: Pubkey) -> Pubkey:
    return Pubkey.find_program_address(
        [b"combined-socialized-loss"],
        program_id,
    )[0]


def get_associated_token_address(authority: Pubkey, mint: Pubkey) -> Pubkey:
    return Pubkey.find_program_address(
        [bytes(authority), bytes(TOKEN_PROGRAM_ID), bytes(mint)],
        ASSOCIATED_TOKEN_PROGRAM_ID,
    )[0]


def get_serum_authority_address(program_id: Pubkey) -> Pubkey:
    return Pubkey.find_program_address(
        [b"serum"],
        program_id,
    )[0]


def get_open_orders_map_address(program_id: Pubkey, open_orders: Pubkey) -> Pubkey:
    return Pubkey.find_program_address(
        [b"cross-open-orders-map", bytes(open_orders)],
        program_id,
    )[0]


def get_mint_authority_address(program_id: Pubkey) -> Pubkey:
    return Pubkey.find_program_address(
        [b"mint-auth"],
        program_id,
    )[0]


def get_underlying_mint_address(asset: Asset, network: Network) -> Pubkey:
    if asset in MINTS:
        return MINTS[asset]
    if asset in FLEXIBLE_MINTS[network]:
        return FLEXIBLE_MINTS[network][asset]
    raise Exception("Underlying mint does not exist!")


def get_referrer_id_account(program_id: Pubkey, id: str) -> Pubkey:
    def remove_trailing_nulls(s):
        return s.rstrip("\x00")

    return Pubkey.find_program_address(
        [b"referrer-id-account", bytes(remove_trailing_nulls(id).encode())],
        program_id,
    )[0]


def get_referrer_pubkey_account(program_id: Pubkey, authority: Pubkey) -> Pubkey:
    return Pubkey.find_program_address(
        [b"referrer-pubkey-account", bytes(authority)],
        program_id,
    )[0]



from __future__ import annotations

import typing

import borsh_construct as borsh
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey

from ..program_id import PROGRAM_ID


class UpdateMakerRebatePercentageArgs(typing.TypedDict):
    native_maker_rebate_percentage: int


layout = borsh.CStruct("native_maker_rebate_percentage" / borsh.U64)


class UpdateMakerRebatePercentageAccounts(typing.TypedDict):
    state: Pubkey
    admin: Pubkey


def update_maker_rebate_percentage(
    args: UpdateMakerRebatePercentageArgs,
    accounts: UpdateMakerRebatePercentageAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["state"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xb4\xec\xfd\x13\xe7\xe7\xdcA"
    encoded_args = layout.build(
        {
            "native_maker_rebate_percentage": args["native_maker_rebate_percentage"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



import collections
import json
import logging
import re
from statistics import median
from typing import List, Optional

import colorlog
from httpx import post
from solana.utils.cluster import cluster_api_url
from solders.hash import Hash

from zetamarkets_py import constants
from zetamarkets_py.types import Asset, Network
from zetamarkets_py.zeta_client.accounts.state import State


def get_fixed_min_lot_size(state: State, asset: Asset) -> int:
    return state.min_lot_sizes[asset.to_index()]


def get_fixed_tick_size(state: State, asset: Asset) -> int:
    return state.tick_sizes[asset.to_index()]


def convert_fixed_int_to_decimal(amount: int) -> float:
    """
    Converts a fixed integer number to a decimal.

    Args:
        amount (int): The fixed integer to be converted.

    Returns:
        float: The converted decimal.
    """
    return amount / 10**constants.PLATFORM_PRECISION


def convert_decimal_to_fixed_int(amount: float, tick_size: int) -> int:
    """
    Converts a decimal to a fixed integer number.

    Args:
        amount (float): The decimal to be converted.

    Returns:
        int: The converted fixed integer.
    """
    return int((amount * 10**constants.PLATFORM_PRECISION / tick_size)) * tick_size


def convert_fixed_lot_to_decimal(amount: int) -> float:
    """
    Converts a fixed integer lot size to a decimal.

    Args:
        amount (int): The fixed lot to be converted.

    Returns:
        float: The converted decimal.
    """
    return amount / 10**constants.POSITION_PRECISION


def convert_decimal_to_fixed_lot(amount: float, min_lot_size: int) -> int:
    """
    Converts a decimal to a fixed integer lot size.

    Args:
        amount (float): The decimal to be converted.

    Returns:
        int: The converted fixed lot.
    """
    return int((amount * 10**constants.POSITION_PRECISION / min_lot_size)) * min_lot_size


def http_to_ws(endpoint: str) -> str:
    """
    Converts an HTTP endpoint to a WebSocket endpoint.

    Args:
        endpoint (str): The HTTP endpoint.

    Returns:
        str: The WebSocket endpoint.
    """
    return re.sub(r"^http", "ws", endpoint)


def cluster_endpoint(network: Network, tls: bool = True, ws: bool = False) -> str:
    """
    Retrieve the RPC API URL for the specified cluster.

    Args:
        network (Network): The network to use.
        tls (bool, optional): If True, use https. Defaults to True.
        ws (bool, optional): If True, use WebSocket. Defaults to False.

    Returns:
        str: The RPC API URL.
    """
    endpoint = cluster_api_url(network.value, tls=tls)  # type: ignore
    if ws:
        ws_endpoint = http_to_ws(endpoint)
        return ws_endpoint
    else:
        return endpoint


def get_tif_offset(expiry_ts: int, epoch_length: int, current_ts: int, tif_buffer: int = 0) -> int:
    """
    Get the Time in Force (TIF) offset.

    Args:
        expiry_ts (int): The expiry timestamp.
        epoch_length (int): The length of the epoch.
        current_ts (int): The current timestamp.

    Returns:
        int: The Time in Force offset.

    Raises:
        Exception: If the expiry timestamp is less than the current timestamp.
    """
    if expiry_ts < current_ts:
        raise Exception(f"Cannot place expired order, current_ts: {current_ts}, expiry_ts: {expiry_ts}")

    # Add tif_buffer here to slow down going into the next epoch to prevent 0x42 errors around epoch
    # The consequence is that you'll send really long expiry orders around the epoch because the tif_offset
    #  will be large even though onchain we've gone to a new epoch
    epoch_start = (current_ts - tif_buffer) - ((current_ts - tif_buffer) % epoch_length)

    tif_offset = expiry_ts - epoch_start
    return min(tif_offset, epoch_length)


def create_logger(name: str, log_level: int = logging.CRITICAL, file_name: Optional[str] = None) -> logging.Logger:
    """
    Create a logger.

    Args:
        name (str): The name of the logger.
        log_level (int, optional): The log level. Defaults to logging.CRITICAL.
        file_name (str, optional): The file name. If provided, logs will be written to this file. Defaults to None.

    Returns:
        logging.Logger: The created logger.
    """
    logger = logging.getLogger(name)
    logger.setLevel(log_level)
    formatter = colorlog.ColoredFormatter("%(log_color)s%(levelname)s:%(name)s:%(message)s")

    # Create console handler and set level to same as the logger
    ch = logging.StreamHandler()
    ch.setLevel(log_level)
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    if file_name:
        fh = logging.FileHandler(file_name)
        fh.setLevel(log_level)
        file_formatter = logging.Formatter("%(asctime)s:%(levelname)s:%(name)s:%(message)s")
        fh.setFormatter(file_formatter)
        logger.addHandler(fh)
    return logger


"""
Fetch the latest priority fee using the getRecentPrioritizationFees RPC method.
This method grabs the minimum fee required to place a tx using a given set of accounts.

Args:
    url (str): The RPC URL to query
    accounts ([str]): List of pubkeys to observe. Zeta market accounts are good for this.
    lookback_slots (int): How many slots to grab the median/max of. Defaults to 20.
    use_max (bool): Whether to use the max fee over the last slots (aggressive). 
        If set to false, the median will be used.
"""


def get_recent_prio_fees(url: str, accounts: List[str], lookback_slots: int = 20, use_max: bool = False) -> int:
    data = {"jsonrpc": "2.0", "id": 1, "method": "getRecentPrioritizationFees", "params": [accounts]}
    headers = {"Content-Type": "application/json"}

    # List of {prioritizationFee, slot}
    response = post(url=url, data=json.dumps(data), headers=headers).json()  # type: ignore

    # Sort by slot, descending. Grab the first 20 slots only
    response_sorted = sorted(response["result"], key=lambda x: x["slot"], reverse=True)[:lookback_slots]

    fees = [value["prioritizationFee"] for value in response_sorted]
    return int(max(fees) if use_max is True else median(fees))


class BlockhashCache:
    """
    A recent blockhash cache that expires after a given number of seconds.
    We grab the oldest cached blockhash in get(), without popping it

    Args:
        ttl: Slots until cached blockhash expires.
    """

    def __init__(self, ttl: int = 60) -> None:
        """Instantiate the cache (you only need to do this once)."""
        self.ttl_slots = 60
        self.blockhashes: collections.deque = collections.deque(maxlen=ttl * 10)

    def set(self, blockhash: Hash, slot: int, used_immediately: bool = False) -> None:
        """
        Update the cache.

        Args:
            blockhash: new blockhash value.
            slot: the slot which the blockhash came from.
            (unused) used_immediately: unused param, exists in the solana BlockhashCache and just guarantees syntax compatibility

        """
        self.blockhashes.append({"slot": slot, "blockhash": blockhash})

        # Keep only N slots of blockhashes
        pops = 0
        for item in self.blockhashes:
            if item["slot"] < slot - self.ttl_slots:
                pops += 1
            else:
                break

        for i in range(pops):
            self.blockhashes.popleft()

    def get(self) -> Hash:
        """
        Get the oldest cached blockhash without popping

        Returns:
            cached blockhash.

        """
        if len(self.blockhashes) > 0:
            return self.blockhashes[0]["blockhash"]
        else:
            raise ValueError



from __future__ import annotations

import typing

import borsh_construct as borsh
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey

from .. import types
from ..program_id import PROGRAM_ID


class UpdateTickSizeArgs(typing.TypedDict):
    asset: types.asset.AssetKind
    tick_size: int


layout = borsh.CStruct("asset" / types.asset.layout, "tick_size" / borsh.U32)


class UpdateTickSizeAccounts(typing.TypedDict):
    state: Pubkey
    admin: Pubkey


def update_tick_size(
    args: UpdateTickSizeArgs,
    accounts: UpdateTickSizeAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["state"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xdez\x01\xdd{t\x8fn"
    encoded_args = layout.build(
        {
            "asset": args["asset"].to_encodable(),
            "tick_size": args["tick_size"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.borsh_extension import EnumForCodegen


class PlaceOrderJSON(typing.TypedDict):
    kind: typing.Literal["PlaceOrder"]


class PlacePerpOrderJSON(typing.TypedDict):
    kind: typing.Literal["PlacePerpOrder"]


@dataclass
class PlaceOrder:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "PlaceOrder"

    @classmethod
    def to_json(cls) -> PlaceOrderJSON:
        return PlaceOrderJSON(
            kind="PlaceOrder",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "PlaceOrder": {},
        }


@dataclass
class PlacePerpOrder:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "PlacePerpOrder"

    @classmethod
    def to_json(cls) -> PlacePerpOrderJSON:
        return PlacePerpOrderJSON(
            kind="PlacePerpOrder",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "PlacePerpOrder": {},
        }


PlaceOrderTypeKind = typing.Union[PlaceOrder, PlacePerpOrder]
PlaceOrderTypeJSON = typing.Union[PlaceOrderJSON, PlacePerpOrderJSON]


def from_decoded(obj: dict) -> PlaceOrderTypeKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "PlaceOrder" in obj:
        return PlaceOrder()
    if "PlacePerpOrder" in obj:
        return PlacePerpOrder()
    raise ValueError("Invalid enum object")


def from_json(obj: PlaceOrderTypeJSON) -> PlaceOrderTypeKind:
    if obj["kind"] == "PlaceOrder":
        return PlaceOrder()
    if obj["kind"] == "PlacePerpOrder":
        return PlacePerpOrder()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen("PlaceOrder" / borsh.CStruct(), "PlacePerpOrder" / borsh.CStruct())



from __future__ import annotations

import typing

import borsh_construct as borsh
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey

from .. import types
from ..program_id import PROGRAM_ID


class ExpireSeriesOverrideArgs(typing.TypedDict):
    args: types.expire_series_override_args.ExpireSeriesOverrideArgs


layout = borsh.CStruct("args" / types.expire_series_override_args.ExpireSeriesOverrideArgs.layout)


def expire_series_override(
    args: ExpireSeriesOverrideArgs,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = []
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b'h\x16"{V\xe0\x82F'
    encoded_args = layout.build(
        {
            "args": args["args"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from .add_market_indexes import AddMarketIndexesAccounts, add_market_indexes
from .add_perp_market_index import (
    AddPerpMarketIndexAccounts,
    AddPerpMarketIndexArgs,
    add_perp_market_index,
)
from .apply_perp_funding import (
    ApplyPerpFundingAccounts,
    ApplyPerpFundingArgs,
    apply_perp_funding,
)
from .burn_vault_tokens import BurnVaultTokensAccounts, burn_vault_tokens
from .cancel_all_market_orders import (
    CancelAllMarketOrdersAccounts,
    CancelAllMarketOrdersArgs,
    cancel_all_market_orders,
)
from .cancel_order import CancelOrderAccounts, CancelOrderArgs, cancel_order
from .cancel_order_by_client_order_id import (
    CancelOrderByClientOrderIdAccounts,
    CancelOrderByClientOrderIdArgs,
    cancel_order_by_client_order_id,
)
from .cancel_order_by_client_order_id_no_error import (
    CancelOrderByClientOrderIdNoErrorAccounts,
    CancelOrderByClientOrderIdNoErrorArgs,
    cancel_order_by_client_order_id_no_error,
)
from .cancel_order_halted import (
    CancelOrderHaltedAccounts,
    CancelOrderHaltedArgs,
    cancel_order_halted,
)
from .cancel_order_no_error import (
    CancelOrderNoErrorAccounts,
    CancelOrderNoErrorArgs,
    cancel_order_no_error,
)
from .cancel_trigger_order import (
    CancelTriggerOrderAccounts,
    CancelTriggerOrderArgs,
    cancel_trigger_order,
)
from .cancel_trigger_order_v2 import (
    CancelTriggerOrderV2Accounts,
    CancelTriggerOrderV2Args,
    cancel_trigger_order_v2,
)
from .choose_airdrop_community import (
    ChooseAirdropCommunityAccounts,
    ChooseAirdropCommunityArgs,
    choose_airdrop_community,
)
from .clean_zeta_market_halted import (
    CleanZetaMarketHaltedAccounts,
    CleanZetaMarketHaltedArgs,
    clean_zeta_market_halted,
)
from .clean_zeta_markets import CleanZetaMarketsAccounts, clean_zeta_markets
from .close_cross_margin_account import (
    CloseCrossMarginAccountAccounts,
    CloseCrossMarginAccountArgs,
    close_cross_margin_account,
)
from .close_cross_margin_account_manager import (
    CloseCrossMarginAccountManagerAccounts,
    close_cross_margin_account_manager,
)
from .close_margin_account import CloseMarginAccountAccounts, close_margin_account
from .close_open_orders import (
    CloseOpenOrdersAccounts,
    CloseOpenOrdersArgs,
    close_open_orders,
)
from .close_open_orders_v2 import (
    CloseOpenOrdersV2Accounts,
    CloseOpenOrdersV2Args,
    close_open_orders_v2,
)
from .close_open_orders_v3 import (
    CloseOpenOrdersV3Accounts,
    CloseOpenOrdersV3Args,
    close_open_orders_v3,
)
from .close_open_orders_v4 import (
    CloseOpenOrdersV4Accounts,
    CloseOpenOrdersV4Args,
    close_open_orders_v4,
)
from .close_referrer_accounts import (
    CloseReferrerAccountsAccounts,
    close_referrer_accounts,
)
from .close_spread_account import CloseSpreadAccountAccounts, close_spread_account
from .collect_treasury_funds import (
    CollectTreasuryFundsAccounts,
    CollectTreasuryFundsArgs,
    collect_treasury_funds,
)
from .crank_event_queue import (
    CrankEventQueueAccounts,
    CrankEventQueueArgs,
    crank_event_queue,
)
from .deposit import DepositAccounts, DepositArgs, deposit
from .deposit_insurance_vault import (
    DepositInsuranceVaultAccounts,
    DepositInsuranceVaultArgs,
    deposit_insurance_vault,
)
from .deposit_insurance_vault_v2 import (
    DepositInsuranceVaultV2Accounts,
    DepositInsuranceVaultV2Args,
    deposit_insurance_vault_v2,
)
from .deposit_permissionless import (
    DepositPermissionlessAccounts,
    DepositPermissionlessArgs,
    deposit_permissionless,
)
from .deposit_v2 import DepositV2Accounts, DepositV2Args, deposit_v2
from .edit_delegated_pubkey import (
    EditDelegatedPubkeyAccounts,
    EditDelegatedPubkeyArgs,
    edit_delegated_pubkey,
)
from .edit_ma_type import EditMaTypeAccounts, EditMaTypeArgs, edit_ma_type
from .edit_trigger_order import (
    EditTriggerOrderAccounts,
    EditTriggerOrderArgs,
    edit_trigger_order,
)
from .edit_trigger_order_v2 import (
    EditTriggerOrderV2Accounts,
    EditTriggerOrderV2Args,
    edit_trigger_order_v2,
)
from .execute_trigger_order import (
    ExecuteTriggerOrderAccounts,
    ExecuteTriggerOrderArgs,
    execute_trigger_order,
)
from .execute_trigger_order_v2 import (
    ExecuteTriggerOrderV2Accounts,
    ExecuteTriggerOrderV2Args,
    execute_trigger_order_v2,
)
from .expire_series import ExpireSeriesArgs, expire_series
from .expire_series_override import ExpireSeriesOverrideArgs, expire_series_override
from .force_cancel_order_by_order_id import (
    ForceCancelOrderByOrderIdAccounts,
    ForceCancelOrderByOrderIdArgs,
    force_cancel_order_by_order_id,
)
from .force_cancel_order_by_order_id_v2 import (
    ForceCancelOrderByOrderIdV2Accounts,
    ForceCancelOrderByOrderIdV2Args,
    force_cancel_order_by_order_id_v2,
)
from .force_cancel_orders import (
    ForceCancelOrdersAccounts,
    ForceCancelOrdersArgs,
    force_cancel_orders,
)
from .force_cancel_orders_v2 import (
    ForceCancelOrdersV2Accounts,
    ForceCancelOrdersV2Args,
    force_cancel_orders_v2,
)
from .force_cancel_trigger_order import (
    ForceCancelTriggerOrderAccounts,
    ForceCancelTriggerOrderArgs,
    force_cancel_trigger_order,
)
from .halt import HaltAccounts, HaltArgs, halt
from .initialize_combined_insurance_vault import (
    InitializeCombinedInsuranceVaultAccounts,
    InitializeCombinedInsuranceVaultArgs,
    initialize_combined_insurance_vault,
)
from .initialize_combined_socialized_loss_account import (
    InitializeCombinedSocializedLossAccountAccounts,
    InitializeCombinedSocializedLossAccountArgs,
    initialize_combined_socialized_loss_account,
)
from .initialize_combined_vault import (
    InitializeCombinedVaultAccounts,
    InitializeCombinedVaultArgs,
    initialize_combined_vault,
)
from .initialize_cross_margin_account import (
    InitializeCrossMarginAccountAccounts,
    InitializeCrossMarginAccountArgs,
    initialize_cross_margin_account,
)
from .initialize_cross_margin_account_manager import (
    InitializeCrossMarginAccountManagerAccounts,
    initialize_cross_margin_account_manager,
)
from .initialize_cross_margin_account_manager_v2 import (
    InitializeCrossMarginAccountManagerV2Accounts,
    InitializeCrossMarginAccountManagerV2Args,
    initialize_cross_margin_account_manager_v2,
)
from .initialize_insurance_deposit_account import (
    InitializeInsuranceDepositAccountAccounts,
    InitializeInsuranceDepositAccountArgs,
    initialize_insurance_deposit_account,
)
from .initialize_margin_account import (
    InitializeMarginAccountAccounts,
    initialize_margin_account,
)
from .initialize_market_indexes import (
    InitializeMarketIndexesAccounts,
    InitializeMarketIndexesArgs,
    initialize_market_indexes,
)
from .initialize_market_node import (
    InitializeMarketNodeAccounts,
    InitializeMarketNodeArgs,
    initialize_market_node,
)
from .initialize_market_strikes import (
    InitializeMarketStrikesAccounts,
    initialize_market_strikes,
)
from .initialize_market_tif_epoch_cycle import (
    InitializeMarketTifEpochCycleAccounts,
    InitializeMarketTifEpochCycleArgs,
    initialize_market_tif_epoch_cycle,
)
from .initialize_min_lots_and_tick_sizes import (
    InitializeMinLotsAndTickSizesAccounts,
    initialize_min_lots_and_tick_sizes,
)
from .initialize_open_orders import InitializeOpenOrdersAccounts, initialize_open_orders
from .initialize_open_orders_v2 import (
    InitializeOpenOrdersV2Accounts,
    initialize_open_orders_v2,
)
from .initialize_open_orders_v3 import (
    InitializeOpenOrdersV3Accounts,
    InitializeOpenOrdersV3Args,
    initialize_open_orders_v3,
)
from .initialize_perp_sync_queue import (
    InitializePerpSyncQueueAccounts,
    InitializePerpSyncQueueArgs,
    initialize_perp_sync_queue,
)
from .initialize_referrer_accounts import (
    InitializeReferrerAccountsAccounts,
    InitializeReferrerAccountsArgs,
    initialize_referrer_accounts,
)
from .initialize_spread_account import (
    InitializeSpreadAccountAccounts,
    initialize_spread_account,
)
from .initialize_underlying import (
    InitializeUnderlyingAccounts,
    InitializeUnderlyingArgs,
    initialize_underlying,
)
from .initialize_whitelist_deposit_account import (
    InitializeWhitelistDepositAccountAccounts,
    InitializeWhitelistDepositAccountArgs,
    initialize_whitelist_deposit_account,
)
from .initialize_whitelist_insurance_account import (
    InitializeWhitelistInsuranceAccountAccounts,
    InitializeWhitelistInsuranceAccountArgs,
    initialize_whitelist_insurance_account,
)
from .initialize_whitelist_trading_fees_account import (
    InitializeWhitelistTradingFeesAccountAccounts,
    InitializeWhitelistTradingFeesAccountArgs,
    initialize_whitelist_trading_fees_account,
)
from .initialize_zeta_group import (
    InitializeZetaGroupAccounts,
    InitializeZetaGroupArgs,
    initialize_zeta_group,
)
from .initialize_zeta_market import (
    InitializeZetaMarketAccounts,
    InitializeZetaMarketArgs,
    initialize_zeta_market,
)
from .initialize_zeta_pricing import (
    InitializeZetaPricingAccounts,
    InitializeZetaPricingArgs,
    initialize_zeta_pricing,
)
from .initialize_zeta_referrals_rewards_wallet import (
    InitializeZetaReferralsRewardsWalletAccounts,
    initialize_zeta_referrals_rewards_wallet,
)
from .initialize_zeta_state import (
    InitializeZetaStateAccounts,
    InitializeZetaStateArgs,
    initialize_zeta_state,
)
from .initialize_zeta_treasury_wallet import (
    InitializeZetaTreasuryWalletAccounts,
    initialize_zeta_treasury_wallet,
)
from .liquidate import LiquidateAccounts, LiquidateArgs, liquidate
from .liquidate_v2 import LiquidateV2Accounts, LiquidateV2Args, liquidate_v2
from .migrate_to_cross_margin_account import (
    MigrateToCrossMarginAccountAccounts,
    migrate_to_cross_margin_account,
)
from .migrate_to_new_cross_margin_account import (
    MigrateToNewCrossMarginAccountAccounts,
    migrate_to_new_cross_margin_account,
)
from .override_expiry import OverrideExpiryAccounts, OverrideExpiryArgs, override_expiry
from .place_multi_orders import (
    PlaceMultiOrdersAccounts,
    PlaceMultiOrdersArgs,
    place_multi_orders,
)
from .place_order import PlaceOrderAccounts, PlaceOrderArgs, place_order
from .place_order_v2 import PlaceOrderV2Accounts, PlaceOrderV2Args, place_order_v2
from .place_order_v3 import PlaceOrderV3Accounts, PlaceOrderV3Args, place_order_v3
from .place_order_v4 import PlaceOrderV4Accounts, PlaceOrderV4Args, place_order_v4
from .place_perp_order import (
    PlacePerpOrderAccounts,
    PlacePerpOrderArgs,
    place_perp_order,
)
from .place_perp_order_v2 import (
    PlacePerpOrderV2Accounts,
    PlacePerpOrderV2Args,
    place_perp_order_v2,
)
from .place_perp_order_v3 import (
    PlacePerpOrderV3Accounts,
    PlacePerpOrderV3Args,
    place_perp_order_v3,
)
from .place_perp_order_v4 import (
    PlacePerpOrderV4Accounts,
    PlacePerpOrderV4Args,
    place_perp_order_v4,
)
from .place_perp_order_v5 import (
    PlacePerpOrderV5Accounts,
    PlacePerpOrderV5Args,
    place_perp_order_v5,
)
from .place_trigger_order import (
    PlaceTriggerOrderAccounts,
    PlaceTriggerOrderArgs,
    place_trigger_order,
)
from .position_movement import (
    PositionMovementAccounts,
    PositionMovementArgs,
    position_movement,
)
from .prune_expired_tif_orders import (
    PruneExpiredTifOrdersAccounts,
    prune_expired_tif_orders,
)
from .prune_expired_tif_orders_v2 import (
    PruneExpiredTifOrdersV2Accounts,
    PruneExpiredTifOrdersV2Args,
    prune_expired_tif_orders_v2,
)
from .rebalance_insurance_vault import (
    RebalanceInsuranceVaultAccounts,
    rebalance_insurance_vault,
)
from .reset_num_flex_underlyings import (
    ResetNumFlexUnderlyingsAccounts,
    reset_num_flex_underlyings,
)
from .settle_dex_funds import SettleDexFundsAccounts, settle_dex_funds
from .settle_positions_halted import (
    SettlePositionsHaltedAccounts,
    SettlePositionsHaltedArgs,
    settle_positions_halted,
)
from .take_trigger_order import (
    TakeTriggerOrderAccounts,
    TakeTriggerOrderArgs,
    take_trigger_order,
)
from .toggle_market_maker import (
    ToggleMarketMakerAccounts,
    ToggleMarketMakerArgs,
    toggle_market_maker,
)
from .toggle_zeta_group_perps_only import (
    ToggleZetaGroupPerpsOnlyAccounts,
    toggle_zeta_group_perps_only,
)
from .transfer_excess_spread_balance import (
    TransferExcessSpreadBalanceAccounts,
    transfer_excess_spread_balance,
)
from .treasury_movement import (
    TreasuryMovementAccounts,
    TreasuryMovementArgs,
    treasury_movement,
)
from .unhalt import UnhaltAccounts, UnhaltArgs, unhalt
from .update_admin import UpdateAdminAccounts, update_admin
from .update_halt_state import (
    UpdateHaltStateAccounts,
    UpdateHaltStateArgs,
    update_halt_state,
)
from .update_interest_rate import (
    UpdateInterestRateAccounts,
    UpdateInterestRateArgs,
    update_interest_rate,
)
from .update_ma_type_admin import UpdateMaTypeAdminAccounts, update_ma_type_admin
from .update_maker_rebate_percentage import (
    UpdateMakerRebatePercentageAccounts,
    UpdateMakerRebatePercentageArgs,
    update_maker_rebate_percentage,
)
from .update_margin_parameters import (
    UpdateMarginParametersAccounts,
    UpdateMarginParametersArgs,
    update_margin_parameters,
)
from .update_min_lot import UpdateMinLotAccounts, UpdateMinLotArgs, update_min_lot
from .update_oracle import UpdateOracleAccounts, update_oracle
from .update_oracle_backup_feed import (
    UpdateOracleBackupFeedAccounts,
    update_oracle_backup_feed,
)
from .update_perp_parameters import (
    UpdatePerpParametersAccounts,
    UpdatePerpParametersArgs,
    update_perp_parameters,
)
from .update_pricing_admin import UpdatePricingAdminAccounts, update_pricing_admin
from .update_pricing_parameters import (
    UpdatePricingParametersAccounts,
    UpdatePricingParametersArgs,
    update_pricing_parameters,
)
from .update_pricing_v2 import (
    UpdatePricingV2Accounts,
    UpdatePricingV2Args,
    update_pricing_v2,
)
from .update_pricing_v3 import (
    UpdatePricingV3Accounts,
    UpdatePricingV3Args,
    update_pricing_v3,
)
from .update_referrals_admin import UpdateReferralsAdminAccounts, update_referrals_admin
from .update_secondary_admin import UpdateSecondaryAdminAccounts, update_secondary_admin
from .update_take_trigger_order_fee_percentage import (
    UpdateTakeTriggerOrderFeePercentageAccounts,
    UpdateTakeTriggerOrderFeePercentageArgs,
    update_take_trigger_order_fee_percentage,
)
from .update_tick_size import (
    UpdateTickSizeAccounts,
    UpdateTickSizeArgs,
    update_tick_size,
)
from .update_trigger_admin import UpdateTriggerAdminAccounts, update_trigger_admin
from .update_volatility import (
    UpdateVolatilityAccounts,
    UpdateVolatilityArgs,
    update_volatility,
)
from .update_zeta_group_expiry_parameters import (
    UpdateZetaGroupExpiryParametersAccounts,
    UpdateZetaGroupExpiryParametersArgs,
    update_zeta_group_expiry_parameters,
)
from .update_zeta_group_margin_parameters import (
    UpdateZetaGroupMarginParametersAccounts,
    UpdateZetaGroupMarginParametersArgs,
    update_zeta_group_margin_parameters,
)
from .update_zeta_group_perp_parameters import (
    UpdateZetaGroupPerpParametersAccounts,
    UpdateZetaGroupPerpParametersArgs,
    update_zeta_group_perp_parameters,
)
from .update_zeta_pricing_pubkeys import (
    UpdateZetaPricingPubkeysAccounts,
    UpdateZetaPricingPubkeysArgs,
    update_zeta_pricing_pubkeys,
)
from .update_zeta_state import (
    UpdateZetaStateAccounts,
    UpdateZetaStateArgs,
    update_zeta_state,
)
from .withdraw import WithdrawAccounts, WithdrawArgs, withdraw
from .withdraw_insurance_vault import (
    WithdrawInsuranceVaultAccounts,
    WithdrawInsuranceVaultArgs,
    withdraw_insurance_vault,
)
from .withdraw_insurance_vault_v2 import (
    WithdrawInsuranceVaultV2Accounts,
    WithdrawInsuranceVaultV2Args,
    withdraw_insurance_vault_v2,
)
from .withdraw_v2 import WithdrawV2Accounts, WithdrawV2Args, withdraw_v2



from __future__ import annotations

import typing

import borsh_construct as borsh
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey

from .. import types
from ..program_id import PROGRAM_ID


class ForceCancelOrdersV2Args(typing.TypedDict):
    asset: types.asset.AssetKind


layout = borsh.CStruct("asset" / types.asset.layout)


class ForceCancelOrdersV2Accounts(typing.TypedDict):
    pricing: Pubkey
    oracle: Pubkey
    oracle_backup_feed: Pubkey
    oracle_backup_program: Pubkey
    cancel_accounts: CancelAccountsNested


class CancelAccountsNested(typing.TypedDict):
    state: Pubkey
    margin_account: Pubkey
    dex_program: Pubkey
    serum_authority: Pubkey
    open_orders: Pubkey
    market: Pubkey
    bids: Pubkey
    asks: Pubkey
    event_queue: Pubkey


def force_cancel_orders_v2(
    args: ForceCancelOrdersV2Args,
    accounts: ForceCancelOrdersV2Accounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["pricing"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["oracle"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["oracle_backup_feed"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["oracle_backup_program"], is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=accounts["cancel_accounts"]["state"],
            is_signer=False,
            is_writable=False,
        ),
        AccountMeta(
            pubkey=accounts["cancel_accounts"]["margin_account"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["cancel_accounts"]["dex_program"],
            is_signer=False,
            is_writable=False,
        ),
        AccountMeta(
            pubkey=accounts["cancel_accounts"]["serum_authority"],
            is_signer=False,
            is_writable=False,
        ),
        AccountMeta(
            pubkey=accounts["cancel_accounts"]["open_orders"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["cancel_accounts"]["market"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["cancel_accounts"]["bids"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["cancel_accounts"]["asks"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["cancel_accounts"]["event_queue"],
            is_signer=False,
            is_writable=True,
        ),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x0e>\x95\xca\x8f\x118s"
    encoded_args = layout.build(
        {
            "asset": args["asset"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing

import borsh_construct as borsh
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID
from solders.sysvar import RENT
from spl.token.constants import TOKEN_PROGRAM_ID

from .. import types
from ..program_id import PROGRAM_ID


class InitializeZetaMarketArgs(typing.TypedDict):
    args: types.initialize_market_args.InitializeMarketArgs


layout = borsh.CStruct("args" / types.initialize_market_args.InitializeMarketArgs.layout)


class InitializeZetaMarketAccounts(typing.TypedDict):
    state: Pubkey
    market_indexes: Pubkey
    pricing: Pubkey
    admin: Pubkey
    market: Pubkey
    request_queue: Pubkey
    event_queue: Pubkey
    bids: Pubkey
    asks: Pubkey
    base_mint: Pubkey
    quote_mint: Pubkey
    zeta_base_vault: Pubkey
    zeta_quote_vault: Pubkey
    dex_base_vault: Pubkey
    dex_quote_vault: Pubkey
    vault_owner: Pubkey
    mint_authority: Pubkey
    serum_authority: Pubkey
    dex_program: Pubkey


def initialize_zeta_market(
    args: InitializeZetaMarketArgs,
    accounts: InitializeZetaMarketAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["state"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["market_indexes"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["pricing"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=accounts["market"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["request_queue"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["event_queue"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["bids"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["asks"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["base_mint"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["quote_mint"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["zeta_base_vault"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["zeta_quote_vault"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["dex_base_vault"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["dex_quote_vault"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["vault_owner"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["mint_authority"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["serum_authority"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["dex_program"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(pubkey=RENT, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"t\xef\xe2\x95.\xa3\xdd\x03"
    encoded_args = layout.build(
        {
            "args": args["args"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing

from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey

from ..program_id import PROGRAM_ID


class CleanZetaMarketsAccounts(typing.TypedDict):
    state: Pubkey
    zeta_group: Pubkey


def clean_zeta_markets(
    accounts: CleanZetaMarketsAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["state"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["zeta_group"], is_signer=False, is_writable=True),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"z\x7f1YD\xe4U\x9d"
    encoded_args = b""
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing

from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey
from spl.token.constants import TOKEN_PROGRAM_ID

from ..program_id import PROGRAM_ID


class RebalanceInsuranceVaultAccounts(typing.TypedDict):
    state: Pubkey
    zeta_vault: Pubkey
    insurance_vault: Pubkey
    treasury_wallet: Pubkey
    socialized_loss_account: Pubkey


def rebalance_insurance_vault(
    accounts: RebalanceInsuranceVaultAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["state"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["zeta_vault"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["insurance_vault"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["treasury_wallet"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["socialized_loss_account"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x0b\xc4B\xeb;\xed\xdfo"
    encoded_args = b""
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.borsh_extension import BorshPubkey
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
from solders.pubkey import Pubkey

from .. import types
from ..program_id import PROGRAM_ID


class GreeksJSON(typing.TypedDict):
    nonce: int
    mark_prices: list[int]
    mark_prices_padding: list[int]
    perp_mark_price: int
    product_greeks: list[types.product_greeks.ProductGreeksJSON]
    product_greeks_padding: list[types.product_greeks.ProductGreeksJSON]
    update_timestamp: list[int]
    update_timestamp_padding: list[int]
    retreat_expiration_timestamp: list[int]
    retreat_expiration_timestamp_padding: list[int]
    interest_rate: list[int]
    interest_rate_padding: list[int]
    nodes: list[int]
    volatility: list[int]
    volatility_padding: list[int]
    node_keys: list[str]
    halt_force_pricing: list[bool]
    perp_update_timestamp: int
    perp_funding_delta: types.anchor_decimal.AnchorDecimalJSON
    perp_latest_funding_rate: types.anchor_decimal.AnchorDecimalJSON
    perp_latest_midpoint: int
    padding: list[int]


@dataclass
class Greeks:
    discriminator: typing.ClassVar = b"\xf7\xd5\xaa\x9a+\xf3\x92\xfe"
    layout: typing.ClassVar = borsh.CStruct(
        "nonce" / borsh.U8,
        "mark_prices" / borsh.U64[46],
        "mark_prices_padding" / borsh.U64[91],
        "perp_mark_price" / borsh.U64,
        "product_greeks" / types.product_greeks.ProductGreeks.layout[22],
        "product_greeks_padding" / types.product_greeks.ProductGreeks.layout[44],
        "update_timestamp" / borsh.U64[2],
        "update_timestamp_padding" / borsh.U64[4],
        "retreat_expiration_timestamp" / borsh.U64[2],
        "retreat_expiration_timestamp_padding" / borsh.U64[4],
        "interest_rate" / borsh.I64[2],
        "interest_rate_padding" / borsh.I64[4],
        "nodes" / borsh.U64[5],
        "volatility" / borsh.U64[10],
        "volatility_padding" / borsh.U64[20],
        "node_keys" / BorshPubkey[138],
        "halt_force_pricing" / borsh.Bool[6],
        "perp_update_timestamp" / borsh.U64,
        "perp_funding_delta" / types.anchor_decimal.AnchorDecimal.layout,
        "perp_latest_funding_rate" / types.anchor_decimal.AnchorDecimal.layout,
        "perp_latest_midpoint" / borsh.U64,
        "padding" / borsh.U8[1593],
    )
    nonce: int
    mark_prices: list[int]
    mark_prices_padding: list[int]
    perp_mark_price: int
    product_greeks: list[types.product_greeks.ProductGreeks]
    product_greeks_padding: list[types.product_greeks.ProductGreeks]
    update_timestamp: list[int]
    update_timestamp_padding: list[int]
    retreat_expiration_timestamp: list[int]
    retreat_expiration_timestamp_padding: list[int]
    interest_rate: list[int]
    interest_rate_padding: list[int]
    nodes: list[int]
    volatility: list[int]
    volatility_padding: list[int]
    node_keys: list[Pubkey]
    halt_force_pricing: list[bool]
    perp_update_timestamp: int
    perp_funding_delta: types.anchor_decimal.AnchorDecimal
    perp_latest_funding_rate: types.anchor_decimal.AnchorDecimal
    perp_latest_midpoint: int
    padding: list[int]

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: Pubkey,
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.Optional["Greeks"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id:
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[Pubkey],
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["Greeks"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["Greeks"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "Greeks":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator("The discriminator for this account is invalid")
        dec = Greeks.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            nonce=dec.nonce,
            mark_prices=dec.mark_prices,
            mark_prices_padding=dec.mark_prices_padding,
            perp_mark_price=dec.perp_mark_price,
            product_greeks=list(
                map(
                    lambda item: types.product_greeks.ProductGreeks.from_decoded(item),
                    dec.product_greeks,
                )
            ),
            product_greeks_padding=list(
                map(
                    lambda item: types.product_greeks.ProductGreeks.from_decoded(item),
                    dec.product_greeks_padding,
                )
            ),
            update_timestamp=dec.update_timestamp,
            update_timestamp_padding=dec.update_timestamp_padding,
            retreat_expiration_timestamp=dec.retreat_expiration_timestamp,
            retreat_expiration_timestamp_padding=dec.retreat_expiration_timestamp_padding,
            interest_rate=dec.interest_rate,
            interest_rate_padding=dec.interest_rate_padding,
            nodes=dec.nodes,
            volatility=dec.volatility,
            volatility_padding=dec.volatility_padding,
            node_keys=dec.node_keys,
            halt_force_pricing=dec.halt_force_pricing,
            perp_update_timestamp=dec.perp_update_timestamp,
            perp_funding_delta=types.anchor_decimal.AnchorDecimal.from_decoded(dec.perp_funding_delta),
            perp_latest_funding_rate=types.anchor_decimal.AnchorDecimal.from_decoded(dec.perp_latest_funding_rate),
            perp_latest_midpoint=dec.perp_latest_midpoint,
            padding=dec.padding,
        )

    def to_json(self) -> GreeksJSON:
        return {
            "nonce": self.nonce,
            "mark_prices": self.mark_prices,
            "mark_prices_padding": self.mark_prices_padding,
            "perp_mark_price": self.perp_mark_price,
            "product_greeks": list(map(lambda item: item.to_json(), self.product_greeks)),
            "product_greeks_padding": list(map(lambda item: item.to_json(), self.product_greeks_padding)),
            "update_timestamp": self.update_timestamp,
            "update_timestamp_padding": self.update_timestamp_padding,
            "retreat_expiration_timestamp": self.retreat_expiration_timestamp,
            "retreat_expiration_timestamp_padding": self.retreat_expiration_timestamp_padding,
            "interest_rate": self.interest_rate,
            "interest_rate_padding": self.interest_rate_padding,
            "nodes": self.nodes,
            "volatility": self.volatility,
            "volatility_padding": self.volatility_padding,
            "node_keys": list(map(lambda item: str(item), self.node_keys)),
            "halt_force_pricing": self.halt_force_pricing,
            "perp_update_timestamp": self.perp_update_timestamp,
            "perp_funding_delta": self.perp_funding_delta.to_json(),
            "perp_latest_funding_rate": self.perp_latest_funding_rate.to_json(),
            "perp_latest_midpoint": self.perp_latest_midpoint,
            "padding": self.padding,
        }

    @classmethod
    def from_json(cls, obj: GreeksJSON) -> "Greeks":
        return cls(
            nonce=obj["nonce"],
            mark_prices=obj["mark_prices"],
            mark_prices_padding=obj["mark_prices_padding"],
            perp_mark_price=obj["perp_mark_price"],
            product_greeks=list(
                map(
                    lambda item: types.product_greeks.ProductGreeks.from_json(item),
                    obj["product_greeks"],
                )
            ),
            product_greeks_padding=list(
                map(
                    lambda item: types.product_greeks.ProductGreeks.from_json(item),
                    obj["product_greeks_padding"],
                )
            ),
            update_timestamp=obj["update_timestamp"],
            update_timestamp_padding=obj["update_timestamp_padding"],
            retreat_expiration_timestamp=obj["retreat_expiration_timestamp"],
            retreat_expiration_timestamp_padding=obj["retreat_expiration_timestamp_padding"],
            interest_rate=obj["interest_rate"],
            interest_rate_padding=obj["interest_rate_padding"],
            nodes=obj["nodes"],
            volatility=obj["volatility"],
            volatility_padding=obj["volatility_padding"],
            node_keys=list(map(lambda item: Pubkey.from_string(item), obj["node_keys"])),
            halt_force_pricing=obj["halt_force_pricing"],
            perp_update_timestamp=obj["perp_update_timestamp"],
            perp_funding_delta=types.anchor_decimal.AnchorDecimal.from_json(obj["perp_funding_delta"]),
            perp_latest_funding_rate=types.anchor_decimal.AnchorDecimal.from_json(obj["perp_latest_funding_rate"]),
            perp_latest_midpoint=obj["perp_latest_midpoint"],
            padding=obj["padding"],
        )



from __future__ import annotations

import typing

import borsh_construct as borsh
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey

from .. import types
from ..program_id import PROGRAM_ID


class CleanZetaMarketHaltedArgs(typing.TypedDict):
    asset: types.asset.AssetKind


layout = borsh.CStruct("asset" / types.asset.layout)


class CleanZetaMarketHaltedAccounts(typing.TypedDict):
    state: Pubkey
    market: Pubkey
    bids: Pubkey
    asks: Pubkey


def clean_zeta_market_halted(
    args: CleanZetaMarketHaltedArgs,
    accounts: CleanZetaMarketHaltedAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["state"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["market"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["bids"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["asks"], is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x89\x8c^\x12\xe7\xe8\xd9\xcc"
    encoded_args = layout.build(
        {
            "asset": args["asset"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing

from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID

from ..program_id import PROGRAM_ID


class InitializeCrossMarginAccountManagerAccounts(typing.TypedDict):
    cross_margin_account_manager: Pubkey
    authority: Pubkey
    payer: Pubkey
    zeta_program: Pubkey


def initialize_cross_margin_account_manager(
    accounts: InitializeCrossMarginAccountManagerAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(
            pubkey=accounts["cross_margin_account_manager"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(pubkey=accounts["authority"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["payer"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=accounts["zeta_program"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"H\x9a\x0f\x1c\xa5\xd7\xd1\xc7"
    encoded_args = b""
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.borsh_extension import BorshPubkey
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
from solders.pubkey import Pubkey

from .. import types
from ..program_id import PROGRAM_ID


class PricingJSON(typing.TypedDict):
    nonce: int
    mark_prices: list[int]
    mark_prices_padding: list[int]
    update_timestamps: list[int]
    update_timestamps_padding: list[int]
    funding_deltas: list[types.anchor_decimal.AnchorDecimalJSON]
    funding_deltas_padding: list[types.anchor_decimal.AnchorDecimalJSON]
    latest_funding_rates: list[types.anchor_decimal.AnchorDecimalJSON]
    latest_funding_rates_padding: list[types.anchor_decimal.AnchorDecimalJSON]
    latest_midpoints: list[int]
    latest_midpoints_padding: list[int]
    oracles: list[str]
    oracles_padding: list[str]
    oracle_backup_feeds: list[str]
    oracle_backup_feeds_padding: list[str]
    markets: list[str]
    markets_padding: list[str]
    perp_sync_queues: list[str]
    perp_sync_queues_padding: list[str]
    perp_parameters: list[types.perp_parameters.PerpParametersJSON]
    perp_parameters_padding: list[types.perp_parameters.PerpParametersJSON]
    margin_parameters: list[types.margin_parameters.MarginParametersJSON]
    margin_parameters_padding: list[types.margin_parameters.MarginParametersJSON]
    products: list[types.product.ProductJSON]
    products_padding: list[types.product.ProductJSON]
    zeta_group_keys: list[str]
    zeta_group_keys_padding: list[str]
    total_insurance_vault_deposits: int
    last_withdraw_timestamp: int
    net_outflow_sum: int
    halt_force_pricing: list[bool]
    halt_force_pricing_padding: list[bool]
    padding: list[int]


@dataclass
class Pricing:
    discriminator: typing.ClassVar = b"\xbe{\xd2\xb6\x8f\x0b\x98\x88"
    layout: typing.ClassVar = borsh.CStruct(
        "nonce" / borsh.U8,
        "mark_prices" / borsh.U64[21],
        "mark_prices_padding" / borsh.U64[4],
        "update_timestamps" / borsh.U64[21],
        "update_timestamps_padding" / borsh.U64[4],
        "funding_deltas" / types.anchor_decimal.AnchorDecimal.layout[21],
        "funding_deltas_padding" / types.anchor_decimal.AnchorDecimal.layout[4],
        "latest_funding_rates" / types.anchor_decimal.AnchorDecimal.layout[21],
        "latest_funding_rates_padding" / types.anchor_decimal.AnchorDecimal.layout[4],
        "latest_midpoints" / borsh.U64[21],
        "latest_midpoints_padding" / borsh.U64[4],
        "oracles" / BorshPubkey[21],
        "oracles_padding" / BorshPubkey[4],
        "oracle_backup_feeds" / BorshPubkey[21],
        "oracle_backup_feeds_padding" / BorshPubkey[4],
        "markets" / BorshPubkey[21],
        "markets_padding" / BorshPubkey[4],
        "perp_sync_queues" / BorshPubkey[21],
        "perp_sync_queues_padding" / BorshPubkey[4],
        "perp_parameters" / types.perp_parameters.PerpParameters.layout[21],
        "perp_parameters_padding" / types.perp_parameters.PerpParameters.layout[4],
        "margin_parameters" / types.margin_parameters.MarginParameters.layout[21],
        "margin_parameters_padding" / types.margin_parameters.MarginParameters.layout[4],
        "products" / types.product.Product.layout[21],
        "products_padding" / types.product.Product.layout[4],
        "zeta_group_keys" / BorshPubkey[21],
        "zeta_group_keys_padding" / BorshPubkey[4],
        "total_insurance_vault_deposits" / borsh.U64,
        "last_withdraw_timestamp" / borsh.U64,
        "net_outflow_sum" / borsh.I64,
        "halt_force_pricing" / borsh.Bool[21],
        "halt_force_pricing_padding" / borsh.Bool[4],
        "padding" / borsh.U8[2707],
    )
    nonce: int
    mark_prices: list[int]
    mark_prices_padding: list[int]
    update_timestamps: list[int]
    update_timestamps_padding: list[int]
    funding_deltas: list[types.anchor_decimal.AnchorDecimal]
    funding_deltas_padding: list[types.anchor_decimal.AnchorDecimal]
    latest_funding_rates: list[types.anchor_decimal.AnchorDecimal]
    latest_funding_rates_padding: list[types.anchor_decimal.AnchorDecimal]
    latest_midpoints: list[int]
    latest_midpoints_padding: list[int]
    oracles: list[Pubkey]
    oracles_padding: list[Pubkey]
    oracle_backup_feeds: list[Pubkey]
    oracle_backup_feeds_padding: list[Pubkey]
    markets: list[Pubkey]
    markets_padding: list[Pubkey]
    perp_sync_queues: list[Pubkey]
    perp_sync_queues_padding: list[Pubkey]
    perp_parameters: list[types.perp_parameters.PerpParameters]
    perp_parameters_padding: list[types.perp_parameters.PerpParameters]
    margin_parameters: list[types.margin_parameters.MarginParameters]
    margin_parameters_padding: list[types.margin_parameters.MarginParameters]
    products: list[types.product.Product]
    products_padding: list[types.product.Product]
    zeta_group_keys: list[Pubkey]
    zeta_group_keys_padding: list[Pubkey]
    total_insurance_vault_deposits: int
    last_withdraw_timestamp: int
    net_outflow_sum: int
    halt_force_pricing: list[bool]
    halt_force_pricing_padding: list[bool]
    padding: list[int]

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: Pubkey,
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.Optional["Pricing"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id:
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[Pubkey],
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["Pricing"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["Pricing"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "Pricing":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator("The discriminator for this account is invalid")
        dec = Pricing.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            nonce=dec.nonce,
            mark_prices=dec.mark_prices,
            mark_prices_padding=dec.mark_prices_padding,
            update_timestamps=dec.update_timestamps,
            update_timestamps_padding=dec.update_timestamps_padding,
            funding_deltas=list(
                map(
                    lambda item: types.anchor_decimal.AnchorDecimal.from_decoded(item),
                    dec.funding_deltas,
                )
            ),
            funding_deltas_padding=list(
                map(
                    lambda item: types.anchor_decimal.AnchorDecimal.from_decoded(item),
                    dec.funding_deltas_padding,
                )
            ),
            latest_funding_rates=list(
                map(
                    lambda item: types.anchor_decimal.AnchorDecimal.from_decoded(item),
                    dec.latest_funding_rates,
                )
            ),
            latest_funding_rates_padding=list(
                map(
                    lambda item: types.anchor_decimal.AnchorDecimal.from_decoded(item),
                    dec.latest_funding_rates_padding,
                )
            ),
            latest_midpoints=dec.latest_midpoints,
            latest_midpoints_padding=dec.latest_midpoints_padding,
            oracles=dec.oracles,
            oracles_padding=dec.oracles_padding,
            oracle_backup_feeds=dec.oracle_backup_feeds,
            oracle_backup_feeds_padding=dec.oracle_backup_feeds_padding,
            markets=dec.markets,
            markets_padding=dec.markets_padding,
            perp_sync_queues=dec.perp_sync_queues,
            perp_sync_queues_padding=dec.perp_sync_queues_padding,
            perp_parameters=list(
                map(
                    lambda item: types.perp_parameters.PerpParameters.from_decoded(item),
                    dec.perp_parameters,
                )
            ),
            perp_parameters_padding=list(
                map(
                    lambda item: types.perp_parameters.PerpParameters.from_decoded(item),
                    dec.perp_parameters_padding,
                )
            ),
            margin_parameters=list(
                map(
                    lambda item: types.margin_parameters.MarginParameters.from_decoded(item),
                    dec.margin_parameters,
                )
            ),
            margin_parameters_padding=list(
                map(
                    lambda item: types.margin_parameters.MarginParameters.from_decoded(item),
                    dec.margin_parameters_padding,
                )
            ),
            products=list(map(lambda item: types.product.Product.from_decoded(item), dec.products)),
            products_padding=list(
                map(
                    lambda item: types.product.Product.from_decoded(item),
                    dec.products_padding,
                )
            ),
            zeta_group_keys=dec.zeta_group_keys,
            zeta_group_keys_padding=dec.zeta_group_keys_padding,
            total_insurance_vault_deposits=dec.total_insurance_vault_deposits,
            last_withdraw_timestamp=dec.last_withdraw_timestamp,
            net_outflow_sum=dec.net_outflow_sum,
            halt_force_pricing=dec.halt_force_pricing,
            halt_force_pricing_padding=dec.halt_force_pricing_padding,
            padding=dec.padding,
        )

    def to_json(self) -> PricingJSON:
        return {
            "nonce": self.nonce,
            "mark_prices": self.mark_prices,
            "mark_prices_padding": self.mark_prices_padding,
            "update_timestamps": self.update_timestamps,
            "update_timestamps_padding": self.update_timestamps_padding,
            "funding_deltas": list(map(lambda item: item.to_json(), self.funding_deltas)),
            "funding_deltas_padding": list(map(lambda item: item.to_json(), self.funding_deltas_padding)),
            "latest_funding_rates": list(map(lambda item: item.to_json(), self.latest_funding_rates)),
            "latest_funding_rates_padding": list(map(lambda item: item.to_json(), self.latest_funding_rates_padding)),
            "latest_midpoints": self.latest_midpoints,
            "latest_midpoints_padding": self.latest_midpoints_padding,
            "oracles": list(map(lambda item: str(item), self.oracles)),
            "oracles_padding": list(map(lambda item: str(item), self.oracles_padding)),
            "oracle_backup_feeds": list(map(lambda item: str(item), self.oracle_backup_feeds)),
            "oracle_backup_feeds_padding": list(map(lambda item: str(item), self.oracle_backup_feeds_padding)),
            "markets": list(map(lambda item: str(item), self.markets)),
            "markets_padding": list(map(lambda item: str(item), self.markets_padding)),
            "perp_sync_queues": list(map(lambda item: str(item), self.perp_sync_queues)),
            "perp_sync_queues_padding": list(map(lambda item: str(item), self.perp_sync_queues_padding)),
            "perp_parameters": list(map(lambda item: item.to_json(), self.perp_parameters)),
            "perp_parameters_padding": list(map(lambda item: item.to_json(), self.perp_parameters_padding)),
            "margin_parameters": list(map(lambda item: item.to_json(), self.margin_parameters)),
            "margin_parameters_padding": list(map(lambda item: item.to_json(), self.margin_parameters_padding)),
            "products": list(map(lambda item: item.to_json(), self.products)),
            "products_padding": list(map(lambda item: item.to_json(), self.products_padding)),
            "zeta_group_keys": list(map(lambda item: str(item), self.zeta_group_keys)),
            "zeta_group_keys_padding": list(map(lambda item: str(item), self.zeta_group_keys_padding)),
            "total_insurance_vault_deposits": self.total_insurance_vault_deposits,
            "last_withdraw_timestamp": self.last_withdraw_timestamp,
            "net_outflow_sum": self.net_outflow_sum,
            "halt_force_pricing": self.halt_force_pricing,
            "halt_force_pricing_padding": self.halt_force_pricing_padding,
            "padding": self.padding,
        }

    @classmethod
    def from_json(cls, obj: PricingJSON) -> "Pricing":
        return cls(
            nonce=obj["nonce"],
            mark_prices=obj["mark_prices"],
            mark_prices_padding=obj["mark_prices_padding"],
            update_timestamps=obj["update_timestamps"],
            update_timestamps_padding=obj["update_timestamps_padding"],
            funding_deltas=list(
                map(
                    lambda item: types.anchor_decimal.AnchorDecimal.from_json(item),
                    obj["funding_deltas"],
                )
            ),
            funding_deltas_padding=list(
                map(
                    lambda item: types.anchor_decimal.AnchorDecimal.from_json(item),
                    obj["funding_deltas_padding"],
                )
            ),
            latest_funding_rates=list(
                map(
                    lambda item: types.anchor_decimal.AnchorDecimal.from_json(item),
                    obj["latest_funding_rates"],
                )
            ),
            latest_funding_rates_padding=list(
                map(
                    lambda item: types.anchor_decimal.AnchorDecimal.from_json(item),
                    obj["latest_funding_rates_padding"],
                )
            ),
            latest_midpoints=obj["latest_midpoints"],
            latest_midpoints_padding=obj["latest_midpoints_padding"],
            oracles=list(map(lambda item: Pubkey.from_string(item), obj["oracles"])),
            oracles_padding=list(map(lambda item: Pubkey.from_string(item), obj["oracles_padding"])),
            oracle_backup_feeds=list(map(lambda item: Pubkey.from_string(item), obj["oracle_backup_feeds"])),
            oracle_backup_feeds_padding=list(
                map(
                    lambda item: Pubkey.from_string(item),
                    obj["oracle_backup_feeds_padding"],
                )
            ),
            markets=list(map(lambda item: Pubkey.from_string(item), obj["markets"])),
            markets_padding=list(map(lambda item: Pubkey.from_string(item), obj["markets_padding"])),
            perp_sync_queues=list(map(lambda item: Pubkey.from_string(item), obj["perp_sync_queues"])),
            perp_sync_queues_padding=list(
                map(
                    lambda item: Pubkey.from_string(item),
                    obj["perp_sync_queues_padding"],
                )
            ),
            perp_parameters=list(
                map(
                    lambda item: types.perp_parameters.PerpParameters.from_json(item),
                    obj["perp_parameters"],
                )
            ),
            perp_parameters_padding=list(
                map(
                    lambda item: types.perp_parameters.PerpParameters.from_json(item),
                    obj["perp_parameters_padding"],
                )
            ),
            margin_parameters=list(
                map(
                    lambda item: types.margin_parameters.MarginParameters.from_json(item),
                    obj["margin_parameters"],
                )
            ),
            margin_parameters_padding=list(
                map(
                    lambda item: types.margin_parameters.MarginParameters.from_json(item),
                    obj["margin_parameters_padding"],
                )
            ),
            products=list(map(lambda item: types.product.Product.from_json(item), obj["products"])),
            products_padding=list(
                map(
                    lambda item: types.product.Product.from_json(item),
                    obj["products_padding"],
                )
            ),
            zeta_group_keys=list(map(lambda item: Pubkey.from_string(item), obj["zeta_group_keys"])),
            zeta_group_keys_padding=list(
                map(
                    lambda item: Pubkey.from_string(item),
                    obj["zeta_group_keys_padding"],
                )
            ),
            total_insurance_vault_deposits=obj["total_insurance_vault_deposits"],
            last_withdraw_timestamp=obj["last_withdraw_timestamp"],
            net_outflow_sum=obj["net_outflow_sum"],
            halt_force_pricing=obj["halt_force_pricing"],
            halt_force_pricing_padding=obj["halt_force_pricing_padding"],
            padding=obj["padding"],
        )



import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.borsh_extension import BorshPubkey
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
from solders.pubkey import Pubkey

from .. import types
from ..program_id import PROGRAM_ID


class TriggerOrderJSON(typing.TypedDict):
    owner: str
    margin_account: str
    open_orders: str
    order_price: int
    trigger_price: typing.Optional[int]
    trigger_ts: typing.Optional[int]
    size: int
    creation_ts: int
    trigger_direction: typing.Optional[types.trigger_direction.TriggerDirectionJSON]
    side: types.side.SideJSON
    asset: types.asset.AssetJSON
    order_type: types.order_type.OrderTypeJSON
    bit: int
    reduce_only: bool


@dataclass
class TriggerOrder:
    discriminator: typing.ClassVar = b"\xec=*\xbe\x98\x0cjt"
    layout: typing.ClassVar = borsh.CStruct(
        "owner" / BorshPubkey,
        "margin_account" / BorshPubkey,
        "open_orders" / BorshPubkey,
        "order_price" / borsh.U64,
        "trigger_price" / borsh.Option(borsh.U64),
        "trigger_ts" / borsh.Option(borsh.U64),
        "size" / borsh.U64,
        "creation_ts" / borsh.U64,
        "trigger_direction" / borsh.Option(types.trigger_direction.layout),
        "side" / types.side.layout,
        "asset" / types.asset.layout,
        "order_type" / types.order_type.layout,
        "bit" / borsh.U8,
        "reduce_only" / borsh.Bool,
    )
    owner: Pubkey
    margin_account: Pubkey
    open_orders: Pubkey
    order_price: int
    trigger_price: typing.Optional[int]
    trigger_ts: typing.Optional[int]
    size: int
    creation_ts: int
    trigger_direction: typing.Optional[types.trigger_direction.TriggerDirectionKind]
    side: types.side.SideKind
    asset: types.asset.AssetKind
    order_type: types.order_type.OrderTypeKind
    bit: int
    reduce_only: bool

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: Pubkey,
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.Optional["TriggerOrder"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id:
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[Pubkey],
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["TriggerOrder"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["TriggerOrder"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "TriggerOrder":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator("The discriminator for this account is invalid")
        dec = TriggerOrder.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            owner=dec.owner,
            margin_account=dec.margin_account,
            open_orders=dec.open_orders,
            order_price=dec.order_price,
            trigger_price=dec.trigger_price,
            trigger_ts=dec.trigger_ts,
            size=dec.size,
            creation_ts=dec.creation_ts,
            trigger_direction=(
                None if dec.trigger_direction is None else types.trigger_direction.from_decoded(dec.trigger_direction)
            ),
            side=types.side.from_decoded(dec.side),
            asset=types.asset.from_decoded(dec.asset),
            order_type=types.order_type.from_decoded(dec.order_type),
            bit=dec.bit,
            reduce_only=dec.reduce_only,
        )

    def to_json(self) -> TriggerOrderJSON:
        return {
            "owner": str(self.owner),
            "margin_account": str(self.margin_account),
            "open_orders": str(self.open_orders),
            "order_price": self.order_price,
            "trigger_price": self.trigger_price,
            "trigger_ts": self.trigger_ts,
            "size": self.size,
            "creation_ts": self.creation_ts,
            "trigger_direction": (None if self.trigger_direction is None else self.trigger_direction.to_json()),
            "side": self.side.to_json(),
            "asset": self.asset.to_json(),
            "order_type": self.order_type.to_json(),
            "bit": self.bit,
            "reduce_only": self.reduce_only,
        }

    @classmethod
    def from_json(cls, obj: TriggerOrderJSON) -> "TriggerOrder":
        return cls(
            owner=Pubkey.from_string(obj["owner"]),
            margin_account=Pubkey.from_string(obj["margin_account"]),
            open_orders=Pubkey.from_string(obj["open_orders"]),
            order_price=obj["order_price"],
            trigger_price=obj["trigger_price"],
            trigger_ts=obj["trigger_ts"],
            size=obj["size"],
            creation_ts=obj["creation_ts"],
            trigger_direction=(
                None
                if obj["trigger_direction"] is None
                else types.trigger_direction.from_json(obj["trigger_direction"])
            ),
            side=types.side.from_json(obj["side"]),
            asset=types.asset.from_json(obj["asset"]),
            order_type=types.order_type.from_json(obj["order_type"]),
            bit=obj["bit"],
            reduce_only=obj["reduce_only"],
        )



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from construct import Container


class ExpireSeriesOverrideArgsJSON(typing.TypedDict):
    settlement_nonce: int
    settlement_price: int


@dataclass
class ExpireSeriesOverrideArgs:
    layout: typing.ClassVar = borsh.CStruct("settlement_nonce" / borsh.U8, "settlement_price" / borsh.U64)
    settlement_nonce: int
    settlement_price: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "ExpireSeriesOverrideArgs":
        return cls(settlement_nonce=obj.settlement_nonce, settlement_price=obj.settlement_price)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "settlement_nonce": self.settlement_nonce,
            "settlement_price": self.settlement_price,
        }

    def to_json(self) -> ExpireSeriesOverrideArgsJSON:
        return {
            "settlement_nonce": self.settlement_nonce,
            "settlement_price": self.settlement_price,
        }

    @classmethod
    def from_json(cls, obj: ExpireSeriesOverrideArgsJSON) -> "ExpireSeriesOverrideArgs":
        return cls(
            settlement_nonce=obj["settlement_nonce"],
            settlement_price=obj["settlement_price"],
        )



from __future__ import annotations

import typing
from dataclasses import dataclass

from construct import BitsInteger, BitsSwapped, BitStruct, Const, Container, Flag


class AccountFlagsJSON(typing.TypedDict):
    initialized: bool
    market: bool
    open_orders: bool
    request_queue: bool
    event_queue: bool
    bids: bool
    asks: bool
    disabled: bool
    closed: bool
    permissioned: bool
    crank_authority_required: bool


@dataclass
class AccountFlags:
    # idk if possible to represent this with borsh
    layout: typing.ClassVar = BitsSwapped(  # swap to little endian
        BitStruct(
            "initialized" / Flag,
            "market" / Flag,
            "open_orders" / Flag,
            "request_queue" / Flag,
            "event_queue" / Flag,
            "bids" / Flag,
            "asks" / Flag,
            "disabled" / Flag,
            "closed" / Flag,
            "permissioned" / Flag,
            "crank_authority_required" / Flag,
            Const(0, BitsInteger(53)),  # padding
        ),
    )
    initialized: bool
    market: bool
    open_orders: bool
    request_queue: bool
    event_queue: bool
    bids: bool
    asks: bool
    disabled: bool
    closed: bool
    permissioned: bool
    crank_authority_required: bool

    @classmethod
    def from_decoded(cls, obj: Container) -> "AccountFlags":
        return cls(
            initialized=obj.initialized,
            market=obj.market,
            open_orders=obj.open_orders,
            request_queue=obj.request_queue,
            event_queue=obj.event_queue,
            bids=obj.bids,
            asks=obj.asks,
            disabled=obj.disabled,
            closed=obj.closed,
            permissioned=obj.permissioned,
            crank_authority_required=obj.crank_authority_required,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "initialized": self.initialized,
            "market": self.market,
            "open_orders": self.open_orders,
            "request_queue": self.request_queue,
            "event_queue": self.event_queue,
            "bids": self.bids,
            "asks": self.asks,
            "disabled": self.disabled,
            "closed": self.closed,
            "permissioned": self.permissioned,
            "crank_authority_required": self.crank_authority_required,
        }

    def to_json(self) -> AccountFlagsJSON:
        return {
            "initialized": self.initialized,
            "market": self.market,
            "open_orders": self.open_orders,
            "request_queue": self.request_queue,
            "event_queue": self.event_queue,
            "bids": self.bids,
            "asks": self.asks,
            "disabled": self.disabled,
            "closed": self.closed,
            "permissioned": self.permissioned,
            "crank_authority_required": self.crank_authority_required,
        }

    @classmethod
    def from_json(cls, obj: AccountFlagsJSON) -> "AccountFlags":
        return cls(
            initialized=obj["initialized"],
            market=obj["market"],
            open_orders=obj["open_orders"],
            request_queue=obj["request_queue"],
            event_queue=obj["event_queue"],
            bids=obj["bids"],
            asks=obj["asks"],
            disabled=obj["disabled"],
            closed=obj["closed"],
            permissioned=obj["permissioned"],
            crank_authority_required=obj["crank_authority_required"],
        )



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from construct import Container


class PerpParametersJSON(typing.TypedDict):
    min_funding_rate_percent: int
    max_funding_rate_percent: int
    impact_cash_delta: int


@dataclass
class PerpParameters:
    layout: typing.ClassVar = borsh.CStruct(
        "min_funding_rate_percent" / borsh.I64,
        "max_funding_rate_percent" / borsh.I64,
        "impact_cash_delta" / borsh.U64,
    )
    min_funding_rate_percent: int
    max_funding_rate_percent: int
    impact_cash_delta: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "PerpParameters":
        return cls(
            min_funding_rate_percent=obj.min_funding_rate_percent,
            max_funding_rate_percent=obj.max_funding_rate_percent,
            impact_cash_delta=obj.impact_cash_delta,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "min_funding_rate_percent": self.min_funding_rate_percent,
            "max_funding_rate_percent": self.max_funding_rate_percent,
            "impact_cash_delta": self.impact_cash_delta,
        }

    def to_json(self) -> PerpParametersJSON:
        return {
            "min_funding_rate_percent": self.min_funding_rate_percent,
            "max_funding_rate_percent": self.max_funding_rate_percent,
            "impact_cash_delta": self.impact_cash_delta,
        }

    @classmethod
    def from_json(cls, obj: PerpParametersJSON) -> "PerpParameters":
        return cls(
            min_funding_rate_percent=obj["min_funding_rate_percent"],
            max_funding_rate_percent=obj["max_funding_rate_percent"],
            impact_cash_delta=obj["impact_cash_delta"],
        )



from __future__ import annotations

import typing

import borsh_construct as borsh
from construct import Construct
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey
from solders.sysvar import RENT
from spl.token.constants import TOKEN_PROGRAM_ID

from .. import types
from ..program_id import PROGRAM_ID


class PlaceMultiOrdersArgs(typing.TypedDict):
    asset: types.asset.AssetKind
    bid_orders: list[types.order_args.OrderArgs]
    ask_orders: list[types.order_args.OrderArgs]
    order_type: types.order_type.OrderTypeKind


layout = borsh.CStruct(
    "asset" / types.asset.layout,
    "bid_orders" / borsh.Vec(typing.cast(Construct, types.order_args.OrderArgs.layout)),
    "ask_orders" / borsh.Vec(typing.cast(Construct, types.order_args.OrderArgs.layout)),
    "order_type" / types.order_type.layout,
)


class PlaceMultiOrdersAccounts(typing.TypedDict):
    authority: Pubkey
    state: Pubkey
    pricing: Pubkey
    margin_account: Pubkey
    dex_program: Pubkey
    serum_authority: Pubkey
    open_orders: Pubkey
    market: Pubkey
    request_queue: Pubkey
    event_queue: Pubkey
    bids: Pubkey
    asks: Pubkey
    market_base_vault: Pubkey
    market_quote_vault: Pubkey
    zeta_base_vault: Pubkey
    zeta_quote_vault: Pubkey
    oracle: Pubkey
    oracle_backup_feed: Pubkey
    oracle_backup_program: Pubkey
    market_base_mint: Pubkey
    market_quote_mint: Pubkey
    mint_authority: Pubkey
    perp_sync_queue: Pubkey


def place_multi_orders(
    args: PlaceMultiOrdersArgs,
    accounts: PlaceMultiOrdersAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["authority"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["state"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["pricing"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["margin_account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["dex_program"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["serum_authority"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["open_orders"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=RENT, is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["market"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["request_queue"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["event_queue"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["bids"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["asks"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["market_base_vault"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["market_quote_vault"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["zeta_base_vault"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["zeta_quote_vault"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["oracle"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["oracle_backup_feed"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["oracle_backup_program"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["market_base_mint"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["market_quote_mint"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["mint_authority"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["perp_sync_queue"], is_signer=False, is_writable=True),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xcc\xd7\xf3\xf3;\xea\xe1y"
    encoded_args = layout.build(
        {
            "asset": args["asset"].to_encodable(),
            "bid_orders": list(map(lambda item: item.to_encodable(), args["bid_orders"])),
            "ask_orders": list(map(lambda item: item.to_encodable(), args["ask_orders"])),
            "order_type": args["order_type"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing

import borsh_construct as borsh
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID

from .. import types
from ..program_id import PROGRAM_ID


class InitializeMarketIndexesArgs(typing.TypedDict):
    nonce: int
    asset: types.asset.AssetKind


layout = borsh.CStruct("nonce" / borsh.U8, "asset" / types.asset.layout)


class InitializeMarketIndexesAccounts(typing.TypedDict):
    state: Pubkey
    market_indexes: Pubkey
    admin: Pubkey
    pricing: Pubkey


def initialize_market_indexes(
    args: InitializeMarketIndexesArgs,
    accounts: InitializeMarketIndexesAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["state"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["market_indexes"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["pricing"], is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"[?\xcd\x90\x14S\xb1x"
    encoded_args = layout.build(
        {
            "nonce": args["nonce"],
            "asset": args["asset"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing

import borsh_construct as borsh
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey
from solders.sysvar import RENT
from spl.token.constants import TOKEN_PROGRAM_ID

from .. import types
from ..program_id import PROGRAM_ID


class PlacePerpOrderV4Args(typing.TypedDict):
    price: int
    size: int
    side: types.side.SideKind
    order_type: types.order_type.OrderTypeKind
    reduce_only: bool
    client_order_id: typing.Optional[int]
    tag: typing.Optional[str]
    tif_offset: typing.Optional[int]
    asset: types.asset.AssetKind


layout = borsh.CStruct(
    "price" / borsh.U64,
    "size" / borsh.U64,
    "side" / types.side.layout,
    "order_type" / types.order_type.layout,
    "reduce_only" / borsh.Bool,
    "client_order_id" / borsh.Option(borsh.U64),
    "tag" / borsh.Option(borsh.String),
    "tif_offset" / borsh.Option(borsh.U16),
    "asset" / types.asset.layout,
)


class PlacePerpOrderV4Accounts(typing.TypedDict):
    authority: Pubkey
    place_order_accounts: PlaceOrderAccountsNested


class PlaceOrderAccountsNested(typing.TypedDict):
    state: Pubkey
    pricing: Pubkey
    margin_account: Pubkey
    dex_program: Pubkey
    serum_authority: Pubkey
    open_orders: Pubkey
    market_accounts: MarketAccountsNested
    oracle: Pubkey
    oracle_backup_feed: Pubkey
    oracle_backup_program: Pubkey
    market_mint: Pubkey
    mint_authority: Pubkey
    perp_sync_queue: Pubkey


class MarketAccountsNested(typing.TypedDict):
    market: Pubkey
    request_queue: Pubkey
    event_queue: Pubkey
    bids: Pubkey
    asks: Pubkey
    order_payer_token_account: Pubkey
    coin_vault: Pubkey
    pc_vault: Pubkey
    coin_wallet: Pubkey
    pc_wallet: Pubkey


def place_perp_order_v4(
    args: PlacePerpOrderV4Args,
    accounts: PlacePerpOrderV4Accounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["authority"], is_signer=True, is_writable=False),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["state"],
            is_signer=False,
            is_writable=False,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["pricing"],
            is_signer=False,
            is_writable=False,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["margin_account"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["dex_program"],
            is_signer=False,
            is_writable=False,
        ),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["serum_authority"],
            is_signer=False,
            is_writable=False,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["open_orders"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(pubkey=RENT, is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["market_accounts"]["market"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["market_accounts"]["request_queue"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["market_accounts"]["event_queue"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["market_accounts"]["bids"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["market_accounts"]["asks"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["market_accounts"]["order_payer_token_account"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["market_accounts"]["coin_vault"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["market_accounts"]["pc_vault"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["market_accounts"]["coin_wallet"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["market_accounts"]["pc_wallet"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["oracle"],
            is_signer=False,
            is_writable=False,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["oracle_backup_feed"],
            is_signer=False,
            is_writable=False,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["oracle_backup_program"],
            is_signer=False,
            is_writable=False,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["market_mint"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["mint_authority"],
            is_signer=False,
            is_writable=False,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["perp_sync_queue"],
            is_signer=False,
            is_writable=True,
        ),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xa6\x8a\xcddZn\xbf["
    encoded_args = layout.build(
        {
            "price": args["price"],
            "size": args["size"],
            "side": args["side"].to_encodable(),
            "order_type": args["order_type"].to_encodable(),
            "reduce_only": args["reduce_only"],
            "client_order_id": args["client_order_id"],
            "tag": args["tag"],
            "tif_offset": args["tif_offset"],
            "asset": args["asset"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing

import borsh_construct as borsh
from construct import Construct
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey

from .. import types
from ..program_id import PROGRAM_ID


class SetReferralsRewardsArgs(typing.TypedDict):
    args: list[types.set_referrals_rewards_args.SetReferralsRewardsArgs]


layout = borsh.CStruct(
    "args" / borsh.Vec(typing.cast(Construct, types.set_referrals_rewards_args.SetReferralsRewardsArgs.layout))
)


class SetReferralsRewardsAccounts(typing.TypedDict):
    state: Pubkey
    referrals_admin: Pubkey


def set_referrals_rewards(
    args: SetReferralsRewardsArgs,
    accounts: SetReferralsRewardsAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["state"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["referrals_admin"], is_signer=True, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b">\xe8\x01\x12K\x07M\x9e"
    encoded_args = layout.build(
        {
            "args": list(map(lambda item: item.to_encodable(), args["args"])),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.borsh_extension import EnumForCodegen


class UndefinedJSON(typing.TypedDict):
    kind: typing.Literal["Undefined"]


class LockJSON(typing.TypedDict):
    kind: typing.Literal["Lock"]


class UnlockJSON(typing.TypedDict):
    kind: typing.Literal["Unlock"]


@dataclass
class Undefined:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Undefined"

    @classmethod
    def to_json(cls) -> UndefinedJSON:
        return UndefinedJSON(
            kind="Undefined",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Undefined": {},
        }


@dataclass
class Lock:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "Lock"

    @classmethod
    def to_json(cls) -> LockJSON:
        return LockJSON(
            kind="Lock",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Lock": {},
        }


@dataclass
class Unlock:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "Unlock"

    @classmethod
    def to_json(cls) -> UnlockJSON:
        return UnlockJSON(
            kind="Unlock",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Unlock": {},
        }


MovementTypeKind = typing.Union[Undefined, Lock, Unlock]
MovementTypeJSON = typing.Union[UndefinedJSON, LockJSON, UnlockJSON]


def from_decoded(obj: dict) -> MovementTypeKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Undefined" in obj:
        return Undefined()
    if "Lock" in obj:
        return Lock()
    if "Unlock" in obj:
        return Unlock()
    raise ValueError("Invalid enum object")


def from_json(obj: MovementTypeJSON) -> MovementTypeKind:
    if obj["kind"] == "Undefined":
        return Undefined()
    if obj["kind"] == "Lock":
        return Lock()
    if obj["kind"] == "Unlock":
        return Unlock()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen("Undefined" / borsh.CStruct(), "Lock" / borsh.CStruct(), "Unlock" / borsh.CStruct())



import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
from solders.pubkey import Pubkey

from .. import types
from ..program_id import PROGRAM_ID


class PerpSyncQueueJSON(typing.TypedDict):
    nonce: int
    head: int
    length: int
    queue: list[types.anchor_decimal.AnchorDecimalJSON]


@dataclass
class PerpSyncQueue:
    discriminator: typing.ClassVar = b"\\78\x9d\xe6\xb8\xabB"
    layout: typing.ClassVar = borsh.CStruct(
        "nonce" / borsh.U8,
        "head" / borsh.U16,
        "length" / borsh.U16,
        "queue" / types.anchor_decimal.AnchorDecimal.layout[600],
    )
    nonce: int
    head: int
    length: int
    queue: list[types.anchor_decimal.AnchorDecimal]

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: Pubkey,
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.Optional["PerpSyncQueue"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id:
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[Pubkey],
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["PerpSyncQueue"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["PerpSyncQueue"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "PerpSyncQueue":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator("The discriminator for this account is invalid")
        dec = PerpSyncQueue.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            nonce=dec.nonce,
            head=dec.head,
            length=dec.length,
            queue=list(
                map(
                    lambda item: types.anchor_decimal.AnchorDecimal.from_decoded(item),
                    dec.queue,
                )
            ),
        )

    def to_json(self) -> PerpSyncQueueJSON:
        return {
            "nonce": self.nonce,
            "head": self.head,
            "length": self.length,
            "queue": list(map(lambda item: item.to_json(), self.queue)),
        }

    @classmethod
    def from_json(cls, obj: PerpSyncQueueJSON) -> "PerpSyncQueue":
        return cls(
            nonce=obj["nonce"],
            head=obj["head"],
            length=obj["length"],
            queue=list(
                map(
                    lambda item: types.anchor_decimal.AnchorDecimal.from_json(item),
                    obj["queue"],
                )
            ),
        )



from __future__ import annotations

import typing

from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID
from solders.sysvar import RENT

from ..program_id import PROGRAM_ID


class InitializeOpenOrdersV2Accounts(typing.TypedDict):
    state: Pubkey
    dex_program: Pubkey
    open_orders: Pubkey
    margin_account: Pubkey
    authority: Pubkey
    payer: Pubkey
    market: Pubkey
    serum_authority: Pubkey
    open_orders_map: Pubkey


def initialize_open_orders_v2(
    accounts: InitializeOpenOrdersV2Accounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["state"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["dex_program"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["open_orders"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["margin_account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["authority"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["payer"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=accounts["market"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["serum_authority"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["open_orders_map"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=RENT, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xdc\x11U\x14p\xae\x94\xe3"
    encoded_args = b""
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from construct import Container


class UpdateGreeksArgsJSON(typing.TypedDict):
    index: int
    theo: int
    delta: int
    gamma: int
    volatility: int


@dataclass
class UpdateGreeksArgs:
    layout: typing.ClassVar = borsh.CStruct(
        "index" / borsh.U8,
        "theo" / borsh.U64,
        "delta" / borsh.U32,
        "gamma" / borsh.U32,
        "volatility" / borsh.U32,
    )
    index: int
    theo: int
    delta: int
    gamma: int
    volatility: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "UpdateGreeksArgs":
        return cls(
            index=obj.index,
            theo=obj.theo,
            delta=obj.delta,
            gamma=obj.gamma,
            volatility=obj.volatility,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "index": self.index,
            "theo": self.theo,
            "delta": self.delta,
            "gamma": self.gamma,
            "volatility": self.volatility,
        }

    def to_json(self) -> UpdateGreeksArgsJSON:
        return {
            "index": self.index,
            "theo": self.theo,
            "delta": self.delta,
            "gamma": self.gamma,
            "volatility": self.volatility,
        }

    @classmethod
    def from_json(cls, obj: UpdateGreeksArgsJSON) -> "UpdateGreeksArgs":
        return cls(
            index=obj["index"],
            theo=obj["theo"],
            delta=obj["delta"],
            gamma=obj["gamma"],
            volatility=obj["volatility"],
        )



import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.borsh_extension import BorshPubkey
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
from solders.pubkey import Pubkey

from .. import types
from ..program_id import PROGRAM_ID


class SpreadAccountJSON(typing.TypedDict):
    authority: str
    nonce: int
    balance: int
    series_expiry: list[int]
    series_expiry_padding: int
    positions: list[types.position.PositionJSON]
    positions_padding: list[types.position.PositionJSON]
    asset: types.asset.AssetJSON
    padding: list[int]


@dataclass
class SpreadAccount:
    discriminator: typing.ClassVar = b"9\x82\xfc\x88\xa7\xb1/\xa2"
    layout: typing.ClassVar = borsh.CStruct(
        "authority" / BorshPubkey,
        "nonce" / borsh.U8,
        "balance" / borsh.U64,
        "series_expiry" / borsh.U64[5],
        "series_expiry_padding" / borsh.U64,
        "positions" / types.position.Position.layout[46],
        "positions_padding" / types.position.Position.layout[92],
        "asset" / types.asset.layout,
        "padding" / borsh.U8[262],
    )
    authority: Pubkey
    nonce: int
    balance: int
    series_expiry: list[int]
    series_expiry_padding: int
    positions: list[types.position.Position]
    positions_padding: list[types.position.Position]
    asset: types.asset.AssetKind
    padding: list[int]

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: Pubkey,
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.Optional["SpreadAccount"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id:
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[Pubkey],
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["SpreadAccount"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["SpreadAccount"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "SpreadAccount":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator("The discriminator for this account is invalid")
        dec = SpreadAccount.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            authority=dec.authority,
            nonce=dec.nonce,
            balance=dec.balance,
            series_expiry=dec.series_expiry,
            series_expiry_padding=dec.series_expiry_padding,
            positions=list(
                map(
                    lambda item: types.position.Position.from_decoded(item),
                    dec.positions,
                )
            ),
            positions_padding=list(
                map(
                    lambda item: types.position.Position.from_decoded(item),
                    dec.positions_padding,
                )
            ),
            asset=types.asset.from_decoded(dec.asset),
            padding=dec.padding,
        )

    def to_json(self) -> SpreadAccountJSON:
        return {
            "authority": str(self.authority),
            "nonce": self.nonce,
            "balance": self.balance,
            "series_expiry": self.series_expiry,
            "series_expiry_padding": self.series_expiry_padding,
            "positions": list(map(lambda item: item.to_json(), self.positions)),
            "positions_padding": list(map(lambda item: item.to_json(), self.positions_padding)),
            "asset": self.asset.to_json(),
            "padding": self.padding,
        }

    @classmethod
    def from_json(cls, obj: SpreadAccountJSON) -> "SpreadAccount":
        return cls(
            authority=Pubkey.from_string(obj["authority"]),
            nonce=obj["nonce"],
            balance=obj["balance"],
            series_expiry=obj["series_expiry"],
            series_expiry_padding=obj["series_expiry_padding"],
            positions=list(
                map(
                    lambda item: types.position.Position.from_json(item),
                    obj["positions"],
                )
            ),
            positions_padding=list(
                map(
                    lambda item: types.position.Position.from_json(item),
                    obj["positions_padding"],
                )
            ),
            asset=types.asset.from_json(obj["asset"]),
            padding=obj["padding"],
        )



from __future__ import annotations

import typing

import borsh_construct as borsh
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID
from solders.sysvar import RENT
from spl.token.constants import TOKEN_PROGRAM_ID

from .. import types
from ..program_id import PROGRAM_ID


class InitializeZetaPricingArgs(typing.TypedDict):
    args: types.initialize_zeta_pricing_args.InitializeZetaPricingArgs


layout = borsh.CStruct("args" / types.initialize_zeta_pricing_args.InitializeZetaPricingArgs.layout)


class InitializeZetaPricingAccounts(typing.TypedDict):
    state: Pubkey
    pricing: Pubkey
    admin: Pubkey


def initialize_zeta_pricing(
    args: InitializeZetaPricingArgs,
    accounts: InitializeZetaPricingAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["state"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["pricing"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=RENT, is_signer=False, is_writable=False),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=True),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"#\xd1\xb4\x1d\xf5\xc7}\x10"
    encoded_args = layout.build(
        {
            "args": args["args"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from construct import Container


class CrossMarginAccountInfoJSON(typing.TypedDict):
    initialized: bool
    name: list[int]


@dataclass
class CrossMarginAccountInfo:
    layout: typing.ClassVar = borsh.CStruct("initialized" / borsh.Bool, "name" / borsh.U8[10])
    initialized: bool
    name: list[int]

    @classmethod
    def from_decoded(cls, obj: Container) -> "CrossMarginAccountInfo":
        return cls(initialized=obj.initialized, name=obj.name)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"initialized": self.initialized, "name": self.name}

    def to_json(self) -> CrossMarginAccountInfoJSON:
        return {"initialized": self.initialized, "name": self.name}

    @classmethod
    def from_json(cls, obj: CrossMarginAccountInfoJSON) -> "CrossMarginAccountInfo":
        return cls(initialized=obj["initialized"], name=obj["name"])



from __future__ import annotations

import typing

import borsh_construct as borsh
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID

from ..program_id import PROGRAM_ID


class InitializeUnderlyingArgs(typing.TypedDict):
    flex_underlying: bool


layout = borsh.CStruct("flex_underlying" / borsh.Bool)


class InitializeUnderlyingAccounts(typing.TypedDict):
    admin: Pubkey
    zeta_program: Pubkey
    state: Pubkey
    underlying: Pubkey
    underlying_mint: Pubkey


def initialize_underlying(
    args: InitializeUnderlyingArgs,
    accounts: InitializeUnderlyingAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["zeta_program"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["state"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["underlying"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["underlying_mint"], is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"rl\xd5\\\xaf|+\x13"
    encoded_args = layout.build(
        {
            "flex_underlying": args["flex_underlying"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing

from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey

from ..program_id import PROGRAM_ID


class PruneExpiredTifOrdersAccounts(typing.TypedDict):
    dex_program: Pubkey
    state: Pubkey
    serum_authority: Pubkey
    market: Pubkey
    bids: Pubkey
    asks: Pubkey
    event_queue: Pubkey


def prune_expired_tif_orders(
    accounts: PruneExpiredTifOrdersAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["dex_program"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["state"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["serum_authority"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["market"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["bids"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["asks"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["event_queue"], is_signer=False, is_writable=True),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x18\xe3\xe2\xd4]\x1a\xf2\xe6"
    encoded_args = b""
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)
